#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the utility functions used by the placement API."""


import datetime
import itertools
from unittest import mock

import fixtures
import microversion_parse
from oslo_middleware import request_id
from oslo_utils.fixture import uuidsentinel
from oslo_utils import timeutils
import testtools
import webob

import placement
from placement import context
from placement import lib as pl
from placement import microversion
from placement.objects import resource_class as rc_obj
from placement.objects import resource_provider as rp_obj
from placement.tests.unit import base
from placement import util
from placement.util import roundrobin


class TestCheckAccept(testtools.TestCase):
    """Confirm behavior of util.check_accept."""

    @staticmethod
    @util.check_accept('application/json', 'application/vnd.openstack')
    def handler(req):
        """Fake handler to test decorator."""
        return True

    def test_fail_no_match(self):
        req = webob.Request.blank('/')
        req.accept = 'text/plain'

        error = self.assertRaises(webob.exc.HTTPNotAcceptable,
                                  self.handler, req)
        self.assertEqual(
            'Only application/json, application/vnd.openstack is provided',
            str(error))

    def test_fail_complex_no_match(self):
        req = webob.Request.blank('/')
        req.accept = 'text/html;q=0.9,text/plain,application/vnd.aws;q=0.8'

        error = self.assertRaises(webob.exc.HTTPNotAcceptable,
                                  self.handler, req)
        self.assertEqual(
            'Only application/json, application/vnd.openstack is provided',
            str(error))

    def test_success_no_accept(self):
        req = webob.Request.blank('/')
        self.assertTrue(self.handler(req))

    def test_success_simple_match(self):
        req = webob.Request.blank('/')
        req.accept = 'application/json'
        self.assertTrue(self.handler(req))

    def test_success_complex_any_match(self):
        req = webob.Request.blank('/')
        req.accept = 'application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'
        self.assertTrue(self.handler(req))

    def test_success_complex_lower_quality_match(self):
        req = webob.Request.blank('/')
        req.accept = 'application/xml;q=0.9,application/vnd.openstack;q=0.8'
        self.assertTrue(self.handler(req))


class TestExtractJSON(testtools.TestCase):

    # Although the intent of this test class is not to test that
    # schemas work, we may as well use a real one to ensure that
    # behaviors are what we expect.
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "uuid": {"type": "string", "format": "uuid"}
        },
        "required": ["name"],
        "additionalProperties": False
    }

    def test_not_json(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  'I am a string',
                                  self.schema)
        self.assertIn('Malformed JSON', str(error))

    def test_malformed_json(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  '{"my bytes got left behind":}',
                                  self.schema)
        self.assertIn('Malformed JSON', str(error))

    def test_schema_mismatch(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  '{"a": "b"}',
                                  self.schema)
        self.assertIn('JSON does not validate', str(error))

    def test_type_invalid(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  '{"name": 1}',
                                  self.schema)
        self.assertIn('JSON does not validate', str(error))

    def test_format_checker(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  '{"name": "hello", "uuid": "not a uuid"}',
                                  self.schema)
        self.assertIn('JSON does not validate', str(error))

    def test_no_additional_properties(self):
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.extract_json,
                                  '{"name": "hello", "cow": "moo"}',
                                  self.schema)
        self.assertIn('JSON does not validate', str(error))

    def test_valid(self):
        data = util.extract_json(
            '{"name": "cow", '
            '"uuid": "%s"}' % uuidsentinel.rp_uuid,
            self.schema)
        self.assertEqual('cow', data['name'])
        self.assertEqual(uuidsentinel.rp_uuid, data['uuid'])


class QueryParamsSchemaTestCase(testtools.TestCase):

    def test_validate_request(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {'type': 'string'}
            },
            'additionalProperties': False}
        req = webob.Request.blank('/test?foo=%88')
        error = self.assertRaises(webob.exc.HTTPBadRequest,
                                  util.validate_query_params,
                                  req, schema)
        self.assertIn('Invalid query string parameters', str(error))


class TestJSONErrorFormatter(testtools.TestCase):

    def setUp(self):
        super(TestJSONErrorFormatter, self).setUp()
        self.environ = {}
        # TODO(jaypipes): Remove this when we get more than a single version
        # in the placement API. The fact that we only had a single version was
        # masking a bug in the utils code.
        _versions = [
            '1.0',
            '1.1',
        ]
        self.useFixture(fixtures.MonkeyPatch('placement.microversion.VERSIONS',
                                             _versions))

    def test_status_to_int_code(self):
        body = ''
        status = '404 Not Found'
        title = ''

        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertEqual(404, result['errors'][0]['status'])

    def test_strip_body_tags(self):
        body = '<h1>Big Error!</h1>'
        status = '400 Bad Request'
        title = ''

        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertEqual('Big Error!', result['errors'][0]['detail'])

    def test_request_id_presence(self):
        body = ''
        status = '400 Bad Request'
        title = ''

        # no request id in environ, none in error
        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertNotIn('request_id', result['errors'][0])

        # request id in environ, request id in error
        self.environ[request_id.ENV_REQUEST_ID] = 'stub-id'

        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertEqual('stub-id', result['errors'][0]['request_id'])

    def test_microversion_406_handling(self):
        body = ''
        status = '400 Bad Request'
        title = ''

        # Not a 406, no version info required.
        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertNotIn('max_version', result['errors'][0])
        self.assertNotIn('min_version', result['errors'][0])

        # A 406 but not because of microversions (microversion
        # parsing was successful), no version info
        # required.
        status = '406 Not Acceptable'
        version_obj = microversion_parse.parse_version_string('2.3')
        self.environ[microversion.MICROVERSION_ENVIRON] = version_obj

        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertNotIn('max_version', result['errors'][0])
        self.assertNotIn('min_version', result['errors'][0])

        # Microversion parsing failed, status is 406, send version info.
        del self.environ[microversion.MICROVERSION_ENVIRON]

        result = util.json_error_formatter(
            body, status, title, self.environ)
        self.assertEqual(microversion.max_version_string(),
                         result['errors'][0]['max_version'])
        self.assertEqual(microversion.min_version_string(),
                         result['errors'][0]['min_version'])


class TestRequireContent(testtools.TestCase):
    """Confirm behavior of util.require_accept."""

    @staticmethod
    @util.require_content('application/json')
    def handler(req):
        """Fake handler to test decorator."""
        return True

    def test_fail_no_content_type(self):
        req = webob.Request.blank('/')

        error = self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                                  self.handler, req)
        self.assertEqual(
            'The media type None is not supported, use application/json',
            str(error))

    def test_fail_wrong_content_type(self):
        req = webob.Request.blank('/')
        req.content_type = 'text/plain'

        error = self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                                  self.handler, req)
        self.assertEqual(
            'The media type text/plain is not supported, use application/json',
            str(error))

    def test_success_content_type(self):
        req = webob.Request.blank('/')
        req.content_type = 'application/json'
        self.assertTrue(self.handler(req))


class TestPlacementURLs(base.ContextTestCase):

    def setUp(self):
        super(TestPlacementURLs, self).setUp()
        fake_context = context.RequestContext(
            user_id='fake', project_id='fake')
        self.resource_provider = rp_obj.ResourceProvider(
            fake_context,
            name=uuidsentinel.rp_name,
            uuid=uuidsentinel.rp_uuid)
        self.resource_class = rc_obj.ResourceClass(
            fake_context,
            name='CUSTOM_BAREMETAL_GOLD',
            id=1000)

    def test_resource_provider_url(self):
        environ = {}
        expected_url = '/resource_providers/%s' % uuidsentinel.rp_uuid
        self.assertEqual(expected_url, util.resource_provider_url(
            environ, self.resource_provider))

    def test_resource_provider_url_prefix(self):
        # SCRIPT_NAME represents the mount point of a WSGI
        # application when it is hosted at a path/prefix.
        environ = {'SCRIPT_NAME': '/placement'}
        expected_url = ('/placement/resource_providers/%s'
                        % uuidsentinel.rp_uuid)
        self.assertEqual(expected_url, util.resource_provider_url(
            environ, self.resource_provider))

    def test_inventories_url(self):
        environ = {}
        expected_url = ('/resource_providers/%s/inventories'
                        % uuidsentinel.rp_uuid)
        self.assertEqual(expected_url, util.inventory_url(
            environ, self.resource_provider))

    def test_inventory_url(self):
        resource_class = 'DISK_GB'
        environ = {}
        expected_url = ('/resource_providers/%s/inventories/%s'
                        % (uuidsentinel.rp_uuid, resource_class))
        self.assertEqual(expected_url, util.inventory_url(
            environ, self.resource_provider, resource_class))

    def test_resource_class_url(self):
        environ = {}
        expected_url = '/resource_classes/CUSTOM_BAREMETAL_GOLD'
        self.assertEqual(expected_url, util.resource_class_url(
            environ, self.resource_class))

    def test_resource_class_url_prefix(self):
        # SCRIPT_NAME represents the mount point of a WSGI
        # application when it is hosted at a path/prefix.
        environ = {'SCRIPT_NAME': '/placement'}
        expected_url = '/placement/resource_classes/CUSTOM_BAREMETAL_GOLD'
        self.assertEqual(expected_url, util.resource_class_url(
            environ, self.resource_class))


class TestNormalizeResourceQsParam(testtools.TestCase):

    def test_success(self):
        qs = "VCPU:1"
        resources = util.normalize_resources_qs_param(qs)
        expected = {
            'VCPU': 1,
        }
        self.assertEqual(expected, resources)

        qs = "VCPU:1,MEMORY_MB:1024,DISK_GB:100"
        resources = util.normalize_resources_qs_param(qs)
        expected = {
            'VCPU': 1,
            'MEMORY_MB': 1024,
            'DISK_GB': 100,
        }
        self.assertEqual(expected, resources)

    def test_400_empty_string(self):
        qs = ""
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_resources_qs_param,
            qs,
        )

    def test_400_bad_int(self):
        qs = "VCPU:foo"
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_resources_qs_param,
            qs,
        )

    def test_400_no_amount(self):
        qs = "VCPU"
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_resources_qs_param,
            qs,
        )

    def test_400_zero_amount(self):
        qs = "VCPU:0"
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_resources_qs_param,
            qs,
        )


class TestNormalizeTraitsQsParamLegacy(testtools.TestCase):

    def test_one(self):
        trait = 'HW_CPU_X86_VMX'
        # Various whitespace permutations
        for fmt in ('%s', ' %s', '%s ', ' %s ', '  %s  '):
            self.assertEqual(
                set([trait]),
                util.normalize_traits_qs_param_to_legacy_value(fmt % trait)
            )

    def test_multiple(self):
        traits = (
            'HW_CPU_X86_VMX',
            'HW_GPU_API_DIRECT3D_V12_0',
            'HW_NIC_OFFLOAD_RX',
            'CUSTOM_GOLD',
            'STORAGE_DISK_SSD',
        )
        self.assertEqual(
            set(traits),
            util.normalize_traits_qs_param_to_legacy_value(
                '%s, %s,%s , %s ,  %s  ' % traits)
        )

    def test_400_all_empty(self):
        for qs in ('', ' ', '   ', ',', ' , , '):
            self.assertRaises(
                webob.exc.HTTPBadRequest,
                util.normalize_traits_qs_param_to_legacy_value, qs)

    def test_400_some_empty(self):
        traits = (
            'HW_NIC_OFFLOAD_RX',
            'CUSTOM_GOLD',
            'STORAGE_DISK_SSD',
        )
        for fmt in ('%s,,%s,%s', ',%s,%s,%s', '%s,%s,%s,', ' %s , %s ,  , %s'):
            self.assertRaises(
                webob.exc.HTTPBadRequest,
                util.normalize_traits_qs_param_to_legacy_value, fmt % traits)


class TestNormalizeTraitsQsParam(testtools.TestCase):

    def test_one(self):
        trait = 'HW_CPU_X86_VMX'
        # Various whitespace permutations
        for fmt in ('%s', ' %s', '%s ', ' %s ', '  %s  '):
            self.assertEqual(
                ([{trait}], set()),
                util.normalize_traits_qs_param(fmt % trait)
            )

    def test_multiple(self):
        traits = (
            'HW_CPU_X86_VMX',
            'HW_GPU_API_DIRECT3D_V12_0',
            'HW_NIC_OFFLOAD_RX',
            'CUSTOM_GOLD',
            'STORAGE_DISK_SSD',
        )
        self.assertEqual(
            ([{trait} for trait in traits], set()),
            util.normalize_traits_qs_param(
                '%s, %s,%s , %s ,  %s  ' % traits)
        )

    def test_400_all_empty(self):
        for qs in ('', ' ', '   ', ',', ' , , '):
            self.assertRaises(
                webob.exc.HTTPBadRequest,
                util.normalize_traits_qs_param, qs)

    def test_400_some_empty(self):
        traits = (
            'HW_NIC_OFFLOAD_RX',
            'CUSTOM_GOLD',
            'STORAGE_DISK_SSD',
        )
        for fmt in (
            '%s,,%s,%s',
            ',%s,%s,%s',
            '%s,%s,%s,',
            ' %s , %s ,  , %s',
            '!,%s,%s,%s',
        ):
            self.assertRaises(
                webob.exc.HTTPBadRequest,
                util.normalize_traits_qs_param,
                fmt % traits,
                allow_forbidden=True,
            )

    def test_multiple_with_forbidden(self):
        traits = (
            '!HW_CPU_X86_VMX',
            'HW_GPU_API_DIRECT3D_V12_0',
            '!HW_NIC_OFFLOAD_RX',
            'CUSTOM_GOLD',
            '!STORAGE_DISK_SSD',
        )

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_traits_qs_param,
            '%s, %s,%s , %s ,  %s  ' % traits,
            allow_forbidden=False)

        self.assertEqual(
            (
                [{'HW_GPU_API_DIRECT3D_V12_0'}, {'CUSTOM_GOLD'}],
                {'HW_CPU_X86_VMX', 'HW_NIC_OFFLOAD_RX', 'STORAGE_DISK_SSD'}),
            util.normalize_traits_qs_param(
                '%s, %s,%s , %s ,  %s  ' % traits, allow_forbidden=True)
        )

    def test_any_traits(self):
        param = 'in:T1 ,T2 , T3'
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_traits_qs_param,
            param,
            allow_any_traits=False
        )
        self.assertEqual(
            ([{'T1', 'T2', 'T3'}], set()),
            util.normalize_traits_qs_param(param, allow_any_traits=True)
        )

    def test_any_traits_not_mix_with_forbidden(self):
        param = 'in:T1 ,!T2 , T3'
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_traits_qs_param,
            param,
            allow_forbidden=True,
            allow_any_traits=True,
        )


class TestNormalizeTraitsQsParams(testtools.TestCase):
    @staticmethod
    def _get_req(qstring, version):
        req = webob.Request.blank(
            '?' + qstring,
        )
        mv_parsed = microversion_parse.Version(*version)
        mv_parsed.max_version = microversion_parse.parse_version_string(
            microversion.max_version_string()
        )
        mv_parsed.min_version = microversion_parse.parse_version_string(
            microversion.min_version_string()
        )
        req.environ[placement.microversion.MICROVERSION_ENVIRON] = mv_parsed
        return req

    def test_suffix(self):
        req = self._get_req('required=!BAZ&requiredX=FOO,BAR', (1, 38))

        required, forbidden = util.normalize_traits_qs_params(req, suffix='')

        self.assertEqual([], required)
        self.assertEqual({'BAZ'}, forbidden)

        required, forbidden = util.normalize_traits_qs_params(req, suffix='X')

        self.assertEqual([{'FOO'}, {'BAR'}], required)
        self.assertEqual(set(), forbidden)

    def test_allow_forbidden_1_21(self):
        req = self._get_req('required=!BAZ', (1, 21))

        ex = self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_traits_qs_params,
            req,
            suffix='',
        )

        self.assertIn(
            "Invalid query string parameters: Expected 'required' parameter "
            "value of the form: HW_CPU_X86_VMX,CUSTOM_MAGIC. Got: !BAZ",
            str(ex),
        )

    def test_allow_forbidden_1_22(self):
        req = self._get_req('required=!BAZ', (1, 22))

        required, forbidden = util.normalize_traits_qs_params(req, suffix='')

        self.assertEqual([], required)
        self.assertEqual({'BAZ'}, forbidden)

    def test_repeated_param_1_38(self):
        req = self._get_req('required=FOO,!BAR&required=BAZ', (1, 38))

        required, forbidden = util.normalize_traits_qs_params(req, suffix='')

        self.assertEqual([{'BAZ'}], required)
        self.assertEqual(set(), forbidden)

    def test_allow_any_traits_1_38(self):
        req = self._get_req('required=in:FOO,BAZ', (1, 38))

        ex = self.assertRaises(
            webob.exc.HTTPBadRequest,
            util.normalize_traits_qs_params,
            req,
            suffix='',
        )

        self.assertIn(
            "Invalid query string parameters: "
            "The format 'in:HW_CPU_X86_VMX,CUSTOM_MAGIC' only supported "
            "since microversion 1.39. Got: in:FOO,BAZ",
            str(ex),
        )

    def test_allow_any_traits_1_39(self):
        req = self._get_req('required=in:FOO,BAZ', (1, 39))

        required, forbidden = util.normalize_traits_qs_params(req, suffix='')

        self.assertEqual([{'FOO', 'BAZ'}], required)
        self.assertEqual(set(), forbidden)

    def test_repeated_param_1_39(self):
        req = self._get_req(
            'required=in:T1,T2'
            '&required=T3,!T4'
            '&required=in:T5,T6'
            '&required=!T7,T8',
            (1, 39)
        )

        required, forbidden = util.normalize_traits_qs_params(req, suffix='')

        self.assertEqual(
            [{'T1', 'T2'}, {'T3'}, {'T5', 'T6'}, {'T8'}],
            required)
        self.assertEqual({'T4', 'T7'}, forbidden)


class TestParseQsRequestGroups(testtools.TestCase):

    @staticmethod
    def do_parse(qstring, version=(1, 18)):
        """Converts a querystring to a MultiDict, mimicking request.GET, and
        runs dict_from_request on it.
        """
        req = webob.Request.blank('?' + qstring)
        mv_parsed = microversion_parse.Version(*version)
        mv_parsed.max_version = microversion_parse.parse_version_string(
            microversion.max_version_string())
        mv_parsed.min_version = microversion_parse.parse_version_string(
            microversion.min_version_string())
        req.environ['placement.microversion'] = mv_parsed
        rqparam = pl.RequestWideParams.from_request(req)
        d = pl.RequestGroup.dict_from_request(req, rqparam)
        # Sort for easier testing
        return [d[suff] for suff in sorted(d)]

    def assertRequestGroupsEqual(self, expected, observed):
        self.assertEqual(len(expected), len(observed))
        for exp, obs in zip(expected, observed):
            self.assertEqual(vars(exp), vars(obs))

    def test_empty_raises(self):
        # TODO(efried): Check the specific error code
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, '')

    def test_unnumbered_only(self):
        """Unnumbered resources & traits - no numbered groupings."""
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&required=HW_CPU_X86_VMX,CUSTOM_GOLD')
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
                required_traits=[
                    {'HW_CPU_X86_VMX'}, {'CUSTOM_GOLD'}
                ],
            ),
        ]
        self.assertRequestGroupsEqual(expected, self.do_parse(qs))

    def test_member_of_single_agg(self):
        """Unnumbered resources with one member_of query param."""
        agg1_uuid = uuidsentinel.agg1
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&member_of=%s' % agg1_uuid)
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
                member_of=[
                    set([agg1_uuid])
                ]
            ),
        ]
        self.assertRequestGroupsEqual(expected, self.do_parse(qs))

    def test_member_of_multiple_aggs_prior_microversion(self):
        """Unnumbered resources with multiple member_of query params before the
        supported microversion should raise a 400.
        """
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&member_of=%s'
              '&member_of=%s' % (agg1_uuid, agg2_uuid))
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_member_of_multiple_aggs(self):
        """Unnumbered resources with multiple member_of query params."""
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&member_of=%s'
              '&member_of=%s' % (agg1_uuid, agg2_uuid))
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
                member_of=[
                    set([agg1_uuid]),
                    set([agg2_uuid])
                ]
            ),
        ]
        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 24)))

    def test_unnumbered_resources_only(self):
        """Validate the bit that can be used for 1.10 and earlier."""
        qs = 'resources=VCPU:2,MEMORY_MB:2048,DISK_GB:5,CUSTOM_MAGIC:123'
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                    'DISK_GB': 5,
                    'CUSTOM_MAGIC': 123,
                },
            ),
        ]
        self.assertRequestGroupsEqual(expected, self.do_parse(qs))

    def test_numbered_only(self):
        # Crazy ordering and nonsequential numbers don't matter.
        # It's okay to have a 'resources' without a 'required'.
        # A trait that's repeated shows up in both spots.
        qs = ('resources1=VCPU:2,MEMORY_MB:2048'
              '&required42=CUSTOM_GOLD'
              '&resources99=DISK_GB:5'
              '&resources42=CUSTOM_MAGIC:123'
              '&required1=HW_CPU_X86_VMX,CUSTOM_GOLD')
        expected = [
            pl.RequestGroup(
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
                required_traits=[
                    {'HW_CPU_X86_VMX'}, {'CUSTOM_GOLD'}
                ],
            ),
            pl.RequestGroup(
                resources={
                    'CUSTOM_MAGIC': 123,
                },
                required_traits=[
                    {'CUSTOM_GOLD'}
                ],
            ),
            pl.RequestGroup(
                resources={
                    'DISK_GB': 5,
                },
            ),
        ]
        self.assertRequestGroupsEqual(expected, self.do_parse(qs))

    def test_numbered_and_unnumbered(self):
        qs = ('resources=VCPU:3,MEMORY_MB:4096,DISK_GB:10'
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&resources1=SRIOV_NET_VF:2'
              '&required1=CUSTOM_PHYSNET_PRIVATE'
              '&resources2=SRIOV_NET_VF:1,NET_INGRESS_BYTES_SEC:20000'
              ',NET_EGRESS_BYTES_SEC:10000'
              '&required2=CUSTOM_SWITCH_BIG,CUSTOM_PHYSNET_PROD'
              '&resources3=CUSTOM_MAGIC:123')
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 3,
                    'MEMORY_MB': 4096,
                    'DISK_GB': 10,
                },
                required_traits=[
                    {'HW_CPU_X86_VMX'},
                    {'CUSTOM_MEM_FLASH'},
                    {'STORAGE_DISK_SSD'}
                ],
            ),
            pl.RequestGroup(
                resources={
                    'SRIOV_NET_VF': 2,
                },
                required_traits=[
                    {'CUSTOM_PHYSNET_PRIVATE'},
                ],
            ),
            pl.RequestGroup(
                resources={
                    'SRIOV_NET_VF': 1,
                    'NET_INGRESS_BYTES_SEC': 20000,
                    'NET_EGRESS_BYTES_SEC': 10000,
                },
                required_traits=[
                    {'CUSTOM_SWITCH_BIG'},
                    {'CUSTOM_PHYSNET_PROD'},
                ],
            ),
            pl.RequestGroup(
                resources={
                    'CUSTOM_MAGIC': 123,
                },
            ),
        ]
        self.assertRequestGroupsEqual(expected, self.do_parse(qs))

    def test_member_of_multiple_aggs_numbered(self):
        """Numbered resources with multiple member_of query params."""
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        agg3_uuid = uuidsentinel.agg3
        agg4_uuid = uuidsentinel.agg4
        qs = ('resources1=VCPU:2'
              '&member_of1=%s'
              '&member_of1=%s'
              '&resources2=VCPU:2'
              '&member_of2=in:%s,%s' % (
                  agg1_uuid, agg2_uuid, agg3_uuid, agg4_uuid))
        expected = [
            pl.RequestGroup(
                resources={
                    'VCPU': 2,
                },
                member_of=[
                    set([agg1_uuid]),
                    set([agg2_uuid])
                ]
            ),
            pl.RequestGroup(
                resources={
                    'VCPU': 2,
                },
                member_of=[
                    set([agg3_uuid, agg4_uuid]),
                ]
            ),
        ]
        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 24)))

    def test_member_of_forbidden_aggs(self):
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        agg3_uuid = uuidsentinel.agg3
        agg4_uuid = uuidsentinel.agg4
        qs = ('resources=VCPU:2'
              '&member_of=%s'
              '&member_of=%s'
              '&member_of=!%s'
              '&member_of=!%s' % (
                  agg1_uuid, agg2_uuid, agg3_uuid, agg4_uuid))
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                },
                member_of=[
                    set([agg1_uuid]),
                    set([agg2_uuid]),
                ],
                forbidden_aggs=set(
                    [agg3_uuid, agg4_uuid]
                ),
            ),
        ]
        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 32)))

    def test_member_of_multiple_forbidden_aggs(self):
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        agg3_uuid = uuidsentinel.agg3
        qs = ('resources=VCPU:2'
              '&member_of=!in:%s,%s,%s' % (
                  agg1_uuid, agg2_uuid, agg3_uuid))
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                },
                forbidden_aggs=set(
                    [agg1_uuid, agg2_uuid, agg3_uuid]
                ),
            ),
        ]
        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 32)))

    def test_member_of_forbidden_aggs_prior_microversion(self):
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        qs = ('resources=VCPU:2'
              '&member_of=!%s'
              '&member_of=!%s' % (agg1_uuid, agg2_uuid))
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 31))

        qs = ('resources=VCPU:2'
              '&member_of=!in:%s,%s' % (agg1_uuid, agg2_uuid))
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 31))

    def test_member_of_forbidden_aggs_invalid_usage(self):
        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        qs = ('resources=VCPU:2'
              '&member_of=in:%s,!%s' % (agg1_uuid, agg2_uuid))
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 32))

        agg1_uuid = uuidsentinel.agg1
        agg2_uuid = uuidsentinel.agg2
        qs = ('resources=VCPU:2'
              '&member_of=!%s,!%s' % (agg1_uuid, agg2_uuid))
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 32))

    def test_400_malformed_resources(self):
        # Somewhat duplicates TestNormalizeResourceQsParam.test_400*.
        qs = ('resources=VCPU:0,MEMORY_MB:4096,DISK_GB:10'
              # Bad ----------^
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&resources1=SRIOV_NET_VF:2'
              '&required1=CUSTOM_PHYSNET_PRIVATE'
              '&resources2=SRIOV_NET_VF:1,NET_INGRESS_BYTES_SEC:20000'
              ',NET_EGRESS_BYTES_SEC:10000'
              '&required2=CUSTOM_SWITCH_BIG,CUSTOM_PHYSNET_PROD'
              '&resources3=CUSTOM_MAGIC:123')
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_400_malformed_traits(self):
        # Somewhat duplicates TestNormalizeResourceQsParam.test_400*.
        qs = ('resources=VCPU:7,MEMORY_MB:4096,DISK_GB:10'
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&resources1=SRIOV_NET_VF:2'
              '&required1=CUSTOM_PHYSNET_PRIVATE'
              '&resources2=SRIOV_NET_VF:1,NET_INGRESS_BYTES_SEC:20000'
              ',NET_EGRESS_BYTES_SEC:10000'
              '&required2=CUSTOM_SWITCH_BIG,CUSTOM_PHYSNET_PROD,'
              # Bad -------------------------------------------^
              '&resources3=CUSTOM_MAGIC:123')
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_400_traits_no_resources_unnumbered(self):
        qs = ('resources9=VCPU:7,MEMORY_MB:4096,DISK_GB:10'
              # Oops ---^
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&resources1=SRIOV_NET_VF:2'
              '&required1=CUSTOM_PHYSNET_PRIVATE'
              '&resources2=SRIOV_NET_VF:1,NET_INGRESS_BYTES_SEC:20000'
              ',NET_EGRESS_BYTES_SEC:10000'
              '&required2=CUSTOM_SWITCH_BIG,CUSTOM_PHYSNET_PROD'
              '&resources3=CUSTOM_MAGIC:123')
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_400_traits_no_resources_numbered(self):
        qs = ('resources=VCPU:7,MEMORY_MB:4096,DISK_GB:10'
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&resources11=SRIOV_NET_VF:2'
              # Oops ----^^
              '&required1=CUSTOM_PHYSNET_PRIVATE'
              '&resources20=SRIOV_NET_VF:1,NET_INGRESS_BYTES_SEC:20000'
              # Oops ----^^
              ',NET_EGRESS_BYTES_SEC:10000'
              '&required2=CUSTOM_SWITCH_BIG,CUSTOM_PHYSNET_PROD'
              '&resources3=CUSTOM_MAGIC:123')
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_400_member_of_no_resources_numbered(self):
        agg1_uuid = uuidsentinel.agg1
        qs = ('resources=VCPU:7,MEMORY_MB:4096,DISK_GB:10'
              '&required=HW_CPU_X86_VMX,CUSTOM_MEM_FLASH,STORAGE_DISK_SSD'
              '&member_of2=%s' % agg1_uuid)
        self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)

    def test_forbidden_one_group(self):
        """When forbidden are allowed this will parse, but otherwise will
        indicate an invalid trait.
        """
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&required=CUSTOM_PHYSNET1,!CUSTOM_SWITCH_BIG')
        expected_forbidden = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
                required_traits=[
                    {'CUSTOM_PHYSNET1'},
                ],
                forbidden_traits={
                    'CUSTOM_SWITCH_BIG',
                }
            ),
        ]
        expected_message = (
            "Invalid query string parameters: Expected 'required' parameter "
            "value of the form: HW_CPU_X86_VMX,CUSTOM_MAGIC. Got: "
            "CUSTOM_PHYSNET1,!CUSTOM_SWITCH_BIG")
        exc = self.assertRaises(webob.exc.HTTPBadRequest, self.do_parse, qs)
        self.assertEqual(expected_message, str(exc))
        self.assertRequestGroupsEqual(
            expected_forbidden, self.do_parse(qs, version=(1, 22)))

    def test_forbidden_conflict(self):
        qs = ('resources=VCPU:2,MEMORY_MB:2048'
              '&required=CUSTOM_PHYSNET1,!CUSTOM_PHYSNET1')

        expected_message = (
            'Conflicting required and forbidden traits found '
            'in the following traits keys: required: (CUSTOM_PHYSNET1)')

        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 22))
        self.assertEqual(expected_message, str(exc))

    def test_forbidden_two_groups(self):
        qs = ('resources=VCPU:2,MEMORY_MB:2048&resources1=CUSTOM_MAGIC:1'
              '&required1=CUSTOM_PHYSNET1,!CUSTOM_PHYSNET2')
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'VCPU': 2,
                    'MEMORY_MB': 2048,
                },
            ),
            pl.RequestGroup(
                resources={
                    'CUSTOM_MAGIC': 1,
                },
                required_traits=[
                    {'CUSTOM_PHYSNET1'},
                ],
                forbidden_traits={
                    'CUSTOM_PHYSNET2',
                }
            ),
        ]

        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 22)))

    def test_forbidden_separate_groups_no_conflict(self):
        qs = ('resources1=CUSTOM_MAGIC:1&required1=CUSTOM_PHYSNET1'
              '&resources2=CUSTOM_MAGIC:1&required2=!CUSTOM_PHYSNET1')
        expected = [
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'CUSTOM_MAGIC': 1,
                },
                required_traits=[
                    {'CUSTOM_PHYSNET1'},
                ],
            ),
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'CUSTOM_MAGIC': 1,
                },
                forbidden_traits={
                    'CUSTOM_PHYSNET1',
                }
            ),
        ]

        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 22)))

    def test_group_suffix_length_1_33(self):
        longstring = '01234567' * 8
        qs = 'resources_%s=CUSTOM_MAGIC:1' % longstring
        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 33))
        # NOTE(cdent): This error message is not what an API user would see.
        # They would get an error during JSON schema processing.
        self.assertIn('least one request group', str(exc))

    def test_group_suffix_character_limits_1_33(self):
        qs = 'resources!#%=CUSTOM_MAGIC:1'
        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 33))
        # NOTE(cdent): This error message is not what an API user would see.
        # They would get an error during JSON schema processing.
        self.assertIn('least one request group', str(exc))

    def test_group_suffix_character_limits_1_22(self):
        qs = 'resources!#%=CUSTOM_MAGIC:1'
        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 22))
        # NOTE(cdent): This error message is not what an API user would see.
        # They would get an error during JSON schema processing.
        self.assertIn('least one request group', str(exc))

    def test_good_suffix_1_33(self):
        qs = ('resources_car_HOUSE_10=CUSTOM_MAGIC:1'
              '&required_car_HOUSE_10=CUSTOM_PHYSNET1')
        expected = [
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'CUSTOM_MAGIC': 1,
                },
                required_traits=[
                    {'CUSTOM_PHYSNET1'},
                ],
            ),
        ]
        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 33)))
        self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 22))

    def test_any_traits_1_38(self):
        qs = 'resources1=RABBIT:1&required1=in:WHITE,BLACK'

        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 38))
        self.assertIn(
            "The format 'in:HW_CPU_X86_VMX,CUSTOM_MAGIC' only supported since "
            "microversion 1.39.",
            str(exc))

    def test_any_traits_1_39(self):
        qs = 'resources1=RABBIT:1&required1=in:WHITE,BLACK'
        expected = [
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'RABBIT': 1,
                },
                required_traits=[
                    {'WHITE', 'BLACK'},
                ],
            ),
        ]

        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 39)))

    def test_any_traits_repeated(self):
        qs = 'resources1=CUSTOM_MAGIC:1&required1=in:T1,T2&required1=T3,!T4'
        expected = [
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'CUSTOM_MAGIC': 1,
                },
                required_traits=[
                    {'T1', 'T2'},
                    {'T3'},
                ],
                forbidden_traits={
                    'T4'
                },
            ),
        ]

        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 39)))

    def test_any_traits_multiple_groups(self):
        qs = ('resources=RABBIT:1&required=in:WHITE,BLACK&'
              'resources2=CAT:2&required2=in:SILVER,RED&required2=!SPOTTED')
        expected = [
            pl.RequestGroup(
                use_same_provider=False,
                resources={
                    'RABBIT': 1,
                },
                required_traits=[
                    {'WHITE', 'BLACK'},
                ],
                forbidden_traits={
                },
            ),
            pl.RequestGroup(
                use_same_provider=True,
                resources={
                    'CAT': 2,
                },
                required_traits=[
                    {'SILVER', 'RED'},
                ],
                forbidden_traits={
                    'SPOTTED'
                },
            ),
        ]

        self.assertRequestGroupsEqual(
            expected, self.do_parse(qs, version=(1, 39)))

    def test_any_traits_forbidden_conflict(self):
        # going against one part of an OR expression is not a conflict as the
        # other parts still can match and fulfill the query
        qs = ('resources=VCPU:2'
              '&required=in:CUSTOM_PHYSNET1,CUSTOM_PHYSNET2'
              '&required=!CUSTOM_PHYSNET1')

        rgs = self.do_parse(qs, version=(1, 39))
        self.assertEqual(1, len(rgs))

        # but going against all parts of an OR expression is a conflict
        qs = ('resources=VCPU:2'
              '&required=in:CUSTOM_PHYSNET1,CUSTOM_PHYSNET2'
              '&required=!CUSTOM_PHYSNET1,!CUSTOM_PHYSNET2')

        expected_message = (
            'Conflicting required and forbidden traits found '
            'in the following traits keys: required: '
            '(CUSTOM_PHYSNET1, CUSTOM_PHYSNET2)')

        exc = self.assertRaises(
            webob.exc.HTTPBadRequest, self.do_parse, qs, version=(1, 39))
        self.assertEqual(expected_message, str(exc))

    def test_stringification(self):
        agg1 = uuidsentinel.agg1
        agg2 = uuidsentinel.agg2
        qs = (f'resources1=CAT:2&required1=in:SILVER,RED&'
              f'required1=TABBY,!SPOTTED&member_of1=in:{agg1},{agg2}')

        rgs = self.do_parse(qs, version=(1, 39))
        self.assertEqual(1, len(rgs))
        self.assertEqual(
            'RequestGroup('
            'use_same_provider=True, '
            'resources={CAT:2}, '
            'traits=((RED or SILVER) and TABBY and !SPOTTED), '
            f'aggregates=[[{", ".join(sorted([agg1, agg2]))}]])',
            str(rgs[0])
        )


class TestPickLastModified(base.ContextTestCase):

    def setUp(self):
        super(TestPickLastModified, self).setUp()
        fake_context = context.RequestContext(
            user_id='fake', project_id='fake')
        self.resource_provider = rp_obj.ResourceProvider(
            fake_context, name=uuidsentinel.rp_name, uuid=uuidsentinel.rp_uuid)

    def test_updated_versus_none(self):
        now = timeutils.utcnow(with_timezone=True)
        self.resource_provider.updated_at = now
        self.resource_provider.created_at = now
        chosen_time = util.pick_last_modified(None, self.resource_provider)
        self.assertEqual(now, chosen_time)

    def test_created_versus_none(self):
        now = timeutils.utcnow(with_timezone=True)
        self.resource_provider.created_at = now
        self.resource_provider.updated_at = None
        chosen_time = util.pick_last_modified(None, self.resource_provider)
        self.assertEqual(now, chosen_time)

    def test_last_modified_less(self):
        now = timeutils.utcnow(with_timezone=True)
        less = now - datetime.timedelta(seconds=300)
        self.resource_provider.updated_at = now
        self.resource_provider.created_at = now
        chosen_time = util.pick_last_modified(less, self.resource_provider)
        self.assertEqual(now, chosen_time)

    def test_last_modified_more(self):
        now = timeutils.utcnow(with_timezone=True)
        more = now + datetime.timedelta(seconds=300)
        self.resource_provider.updated_at = now
        self.resource_provider.created_at = now
        chosen_time = util.pick_last_modified(more, self.resource_provider)
        self.assertEqual(more, chosen_time)

    def test_last_modified_same(self):
        now = timeutils.utcnow(with_timezone=True)
        self.resource_provider.updated_at = now
        self.resource_provider.created_at = now
        chosen_time = util.pick_last_modified(now, self.resource_provider)
        self.assertEqual(now, chosen_time)

    def test_no_object_time_fields_less(self):
        # An unsaved ovo will not have the created_at or updated_at fields
        # present on the object at all.
        now = timeutils.utcnow(with_timezone=True)
        less = now - datetime.timedelta(seconds=300)
        with mock.patch('oslo_utils.timeutils.utcnow') as mock_utc:
            mock_utc.return_value = now
            chosen_time = util.pick_last_modified(
                less, self.resource_provider)
            self.assertEqual(now, chosen_time)
            mock_utc.assert_called_once_with(with_timezone=True)

    def test_no_object_time_fields_more(self):
        # An unsaved ovo will not have the created_at or updated_at fields
        # present on the object at all.
        now = timeutils.utcnow(with_timezone=True)
        more = now + datetime.timedelta(seconds=300)
        with mock.patch('oslo_utils.timeutils.utcnow') as mock_utc:
            mock_utc.return_value = now
            chosen_time = util.pick_last_modified(
                more, self.resource_provider)
            self.assertEqual(more, chosen_time)
            mock_utc.assert_called_once_with(with_timezone=True)

    def test_no_object_time_fields_none(self):
        # An unsaved ovo will not have the created_at or updated_at fields
        # present on the object at all.
        now = timeutils.utcnow(with_timezone=True)
        with mock.patch('oslo_utils.timeutils.utcnow') as mock_utc:
            mock_utc.return_value = now
            chosen_time = util.pick_last_modified(
                None, self.resource_provider)
            self.assertEqual(now, chosen_time)
            mock_utc.assert_called_once_with(with_timezone=True)


class RunOnceTests(testtools.TestCase):

    fake_logger = mock.MagicMock()

    @util.run_once("already ran once", fake_logger)
    def dummy_test_func(self, fail=False):
        if fail:
            raise ValueError()
        return True

    def setUp(self):
        super(RunOnceTests, self).setUp()
        self.dummy_test_func.reset()
        RunOnceTests.fake_logger.reset_mock()

    def test_wrapped_funtions_called_once(self):
        self.assertFalse(self.dummy_test_func.called)
        result = self.dummy_test_func()
        self.assertTrue(result)
        self.assertTrue(self.dummy_test_func.called)

        # assert that on second invocation no result
        # is returned and that the logger is invoked.
        result = self.dummy_test_func()
        RunOnceTests.fake_logger.assert_called_once()
        self.assertIsNone(result)

    def test_wrapped_funtions_called_once_raises(self):
        self.assertFalse(self.dummy_test_func.called)
        self.assertRaises(ValueError, self.dummy_test_func, fail=True)
        self.assertTrue(self.dummy_test_func.called)

        # assert that on second invocation no result
        # is returned and that the logger is invoked.
        result = self.dummy_test_func()
        RunOnceTests.fake_logger.assert_called_once()
        self.assertIsNone(result)

    def test_wrapped_funtions_can_be_reset(self):
        # assert we start with a clean state
        self.assertFalse(self.dummy_test_func.called)
        result = self.dummy_test_func()
        self.assertTrue(result)

        self.dummy_test_func.reset()
        # assert we restored a clean state
        self.assertFalse(self.dummy_test_func.called)
        result = self.dummy_test_func()
        self.assertTrue(result)

        # assert that we never called the logger
        RunOnceTests.fake_logger.assert_not_called()

    def test_reset_calls_cleanup(self):
        mock_clean = mock.Mock()

        @util.run_once("already ran once", self.fake_logger,
                       cleanup=mock_clean)
        def f():
            pass

        f()
        self.assertTrue(f.called)

        f.reset()
        self.assertFalse(f.called)
        mock_clean.assert_called_once_with()

    def test_clean_is_not_called_at_reset_if_wrapped_not_called(self):
        mock_clean = mock.Mock()

        @util.run_once("already ran once", self.fake_logger,
                       cleanup=mock_clean)
        def f():
            pass

        self.assertFalse(f.called)

        f.reset()
        self.assertFalse(f.called)
        self.assertFalse(mock_clean.called)

    def test_reset_works_even_if_cleanup_raises(self):
        mock_clean = mock.Mock(side_effect=ValueError())

        @util.run_once("already ran once", self.fake_logger,
                       cleanup=mock_clean)
        def f():
            pass

        f()
        self.assertTrue(f.called)

        self.assertRaises(ValueError, f.reset)
        self.assertFalse(f.called)
        mock_clean.assert_called_once_with()


class RoundRobinTests(testtools.TestCase):
    def test_no_input(self):
        self.assertEqual([], list(roundrobin()))

    def test_single_input(self):
        self.assertEqual([1, 2], list(roundrobin(iter([1, 2]))))

    def test_balanced_inputs(self):
        self.assertEqual(
            [1, "x", 2, "y"],
            list(roundrobin(
                iter([1, 2]),
                iter(["x", "y"]))
            )
        )

    def test_unbalanced_inputs(self):
        self.assertEqual(
            ["A", "D", "E", "B", "F", "C"],
            list(roundrobin(
                iter("ABC"),
                iter("D"),
                iter("EF"))
            )
        )


class ProductGeneratorTest(testtools.TestCase):
    @staticmethod
    def _no_skip(*args, **kwargs):
        return False

    def test_no_input(self):
        self.assertEqual(
            [()], list(util.filtered_product(should_skip=self._no_skip)))

    def test_full_product_no_pruning(self):
        product = list(util.filtered_product(self._no_skip, [0, 1], [0, 1]))
        self.assertEqual(4, len(product))
        self.assertEqual([(0, 0), (0, 1), (1, 0), (1, 1)], product)

    def test_product_with_filtering(self):
        """This test shows that the filtering removes products from the
        results that are invalid
        """

        def reject_if_non_unique_items(partial_product):
            # A simple check that rejects a partial candidate if any of the
            # items in the candidate are non-unique
            return len(set(partial_product)) != len(partial_product)

        product = list(util.filtered_product(
            reject_if_non_unique_items, [0, 1], [0, 1]))
        self.assertEqual(2, len(product))
        self.assertEqual([(0, 1), (1, 0)], product)

    def test_product_with_pruning(self):
        """This test shows that if the filter rejects partial candidates then
        the product generation will skip generating any product with that
        partial prefix and therefore takes fewer steps than the same
        itertools.product() call.
        """
        mock_filter = mock.Mock(return_value=False)
        product = list(
            util.filtered_product(mock_filter, *([range(3)] * 3)))
        self.assertEqual(27, len(product))
        self.assertEqual(product, list(itertools.product(*([range(3)] * 3))))

        nr_of_check_calls = 0

        def reject_if_non_unique_items(partial_product):
            # A simple check that rejects a partial candidate if any of the
            # items in the candidate are non-unique. While counting the
            # number of times the check is called.
            nonlocal nr_of_check_calls
            nr_of_check_calls += 1
            return len(set(partial_product)) != len(partial_product)

        product = list(util.filtered_product(
            reject_if_non_unique_items, *([range(3)] * 3)))
        self.assertEqual(6, len(product))
        self.assertEqual(
            [(0, 1, 2),
             (0, 2, 1),
             (1, 0, 2),
             (1, 2, 0),
             (2, 0, 1),
             (2, 1, 0)],
            product)

        # If a filter is provided that rejected partial candidates then the
        # product generation does fewer actual steps, this is the pruning we
        # want
        self.assertLess(nr_of_check_calls, len(mock_filter.mock_calls))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from placement.policies import base


RP_ALLOC_LIST = 'placement:resource_providers:allocations:list'

ALLOC_PREFIX = 'placement:allocations:%s'
ALLOC_LIST = ALLOC_PREFIX % 'list'
ALLOC_MANAGE = ALLOC_PREFIX % 'manage'
ALLOC_UPDATE = ALLOC_PREFIX % 'update'
ALLOC_DELETE = ALLOC_PREFIX % 'delete'

rules = [
    policy.DocumentedRuleDefault(
        name=ALLOC_MANAGE,
        check_str=base.ADMIN_OR_SERVICE,
        description="Manage allocations.",
        operations=[
            {
                'method': 'POST',
                'path': '/allocations'
            }
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        name=ALLOC_LIST,
        check_str=base.ADMIN_OR_SERVICE,
        description="List allocations.",
        operations=[
            {
                'method': 'GET',
                'path': '/allocations/{consumer_uuid}'
            }
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        name=ALLOC_UPDATE,
        check_str=base.ADMIN_OR_SERVICE,
        description="Update allocations.",
        operations=[
            {
                'method': 'PUT',
                'path': '/allocations/{consumer_uuid}'
            }
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        name=ALLOC_DELETE,
        check_str=base.ADMIN_OR_SERVICE,
        description="Delete allocations.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/allocations/{consumer_uuid}'
            }
        ],
        scope_types=['project'],
    ),
    policy.DocumentedRuleDefault(
        name=RP_ALLOC_LIST,
        check_str=base.ADMIN_OR_SERVICE,
        description="List resource provider allocations.",
        operations=[
            {
                'method': 'GET',
                'path': '/resource_providers/{uuid}/allocations'
            }
        ],
        scope_types=['project'],
    ),
]


def list_rules():
    return rules



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

api_group = cfg.OptGroup(
    'api',
    title='API options',
    help="""
Options under this group are used to define Placement API.
""")

api_opts = [
    cfg.StrOpt(
        "auth_strategy",
        default="keystone",
        choices=("keystone", "noauth2"),
        deprecated_group="DEFAULT",
        help="""
This determines the strategy to use for authentication: keystone or noauth2.
'noauth2' is designed for testing only, as it does no actual credential
checking. 'noauth2' provides administrative credentials only if 'admin' is
specified as the username.
"""),
]


def register_opts(conf):
    conf.register_group(api_group)
    conf.register_opts(api_opts, group=api_group)


def list_opts():
    return {api_group: api_opts}



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the utility functions used by the placement DB."""

import fixtures
import microversion_parse
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_utils.fixture import uuidsentinel
import webob

from placement import conf
from placement import context
from placement import exception
from placement.handlers import util
from placement import microversion
from placement.objects import consumer as consumer_obj
from placement.objects import project as project_obj
from placement.objects import user as user_obj
from placement.tests.unit import base


class TestEnsureConsumer(base.ContextTestCase):
    def setUp(self):
        super(TestEnsureConsumer, self).setUp()
        self.conf = cfg.ConfigOpts()
        self.useFixture(config_fixture.Config(self.conf))
        conf.register_opts(self.conf)
        self.mock_project_get = self.useFixture(fixtures.MockPatch(
            'placement.objects.project.'
            'Project.get_by_external_id')).mock
        self.mock_user_get = self.useFixture(fixtures.MockPatch(
            'placement.objects.user.'
            'User.get_by_external_id')).mock
        self.mock_consumer_get = self.useFixture(fixtures.MockPatch(
            'placement.objects.consumer.'
            'Consumer.get_by_uuid')).mock
        self.mock_project_create = self.useFixture(fixtures.MockPatch(
            'placement.objects.project.'
            'Project.create')).mock
        self.mock_user_create = self.useFixture(fixtures.MockPatch(
            'placement.objects.user.'
            'User.create')).mock
        self.mock_consumer_create = self.useFixture(fixtures.MockPatch(
            'placement.objects.consumer.'
            'Consumer.create')).mock
        self.mock_consumer_update = self.useFixture(fixtures.MockPatch(
            'placement.objects.consumer.'
            'Consumer.update')).mock
        self.ctx = context.RequestContext(user_id='fake', project_id='fake')
        self.ctx.config = self.conf
        self.consumer_id = uuidsentinel.consumer
        self.project_id = uuidsentinel.project
        self.user_id = uuidsentinel.user
        mv_parsed = microversion_parse.Version(1, 27)
        mv_parsed.max_version = microversion_parse.parse_version_string(
            microversion.max_version_string())
        mv_parsed.min_version = microversion_parse.parse_version_string(
            microversion.min_version_string())
        self.before_version = mv_parsed
        mv_parsed = microversion_parse.Version(1, 28)
        mv_parsed.max_version = microversion_parse.parse_version_string(
            microversion.max_version_string())
        mv_parsed.min_version = microversion_parse.parse_version_string(
            microversion.min_version_string())
        self.after_version = mv_parsed
        mv_parsed = microversion_parse.Version(1, 38)
        mv_parsed.max_version = microversion_parse.parse_version_string(
            microversion.max_version_string())
        mv_parsed.min_version = microversion_parse.parse_version_string(
            microversion.min_version_string())
        self.cons_type_req_version = mv_parsed

    def test_no_existing_project_user_consumer_before_gen_success(self):
        """Tests that we don't require a consumer_generation=None before the
        appropriate microversion.
        """
        self.mock_project_get.side_effect = exception.NotFound
        self.mock_user_get.side_effect = exception.NotFound
        self.mock_consumer_get.side_effect = exception.NotFound

        consumer_gen = 1  # should be ignored
        util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.before_version)

        self.mock_project_get.assert_called_once_with(
            self.ctx, self.project_id)
        self.mock_user_get.assert_called_once_with(
            self.ctx, self.user_id)
        self.mock_consumer_get.assert_called_once_with(
            self.ctx, self.consumer_id)
        self.mock_project_create.assert_called_once()
        self.mock_user_create.assert_called_once()
        self.mock_consumer_create.assert_called_once()

    def test_no_existing_project_user_consumer_after_gen_success(self):
        """Tests that we require a consumer_generation=None after the
        appropriate microversion.
        """
        self.mock_project_get.side_effect = exception.NotFound
        self.mock_user_get.side_effect = exception.NotFound
        self.mock_consumer_get.side_effect = exception.NotFound

        consumer_gen = None  # should NOT be ignored (and None is expected)
        util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.after_version)

        self.mock_project_get.assert_called_once_with(
            self.ctx, self.project_id)
        self.mock_user_get.assert_called_once_with(
            self.ctx, self.user_id)
        self.mock_consumer_get.assert_called_once_with(
            self.ctx, self.consumer_id)
        self.mock_project_create.assert_called_once()
        self.mock_user_create.assert_called_once()
        self.mock_consumer_create.assert_called_once()

    def test_no_existing_project_user_consumer_after_gen_fail(self):
        """Tests that we require a consumer_generation=None after the
        appropriate microversion and that None is the expected value.
        """
        self.mock_project_get.side_effect = exception.NotFound
        self.mock_user_get.side_effect = exception.NotFound
        self.mock_consumer_get.side_effect = exception.NotFound

        consumer_gen = 1  # should NOT be ignored (and 1 is not expected)
        self.assertRaises(
            webob.exc.HTTPConflict,
            util.ensure_consumer,
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.after_version)

    def test_no_existing_project_user_consumer_use_incomplete(self):
        """Verify that if the project_id arg is None, that we fall back to the
        CONF options for incomplete project and user ID.
        """
        self.mock_project_get.side_effect = exception.NotFound
        self.mock_user_get.side_effect = exception.NotFound
        self.mock_consumer_get.side_effect = exception.NotFound

        consumer_gen = None  # should NOT be ignored (and None is expected)
        util.ensure_consumer(
            self.ctx, self.consumer_id, None, None,
            consumer_gen, 'TYPE', self.before_version)

        self.mock_project_get.assert_called_once_with(
            self.ctx, self.conf.placement.incomplete_consumer_project_id)
        self.mock_user_get.assert_called_once_with(
            self.ctx, self.conf.placement.incomplete_consumer_user_id)
        self.mock_consumer_get.assert_called_once_with(
            self.ctx, self.consumer_id)
        self.mock_project_create.assert_called_once()
        self.mock_user_create.assert_called_once()
        self.mock_consumer_create.assert_called_once()

    def test_existing_project_no_existing_consumer_before_gen_success(self):
        """Check that if we find an existing project and user, that we use
        those found objects in creating the consumer. Do not require a consumer
        generation before the appropriate microversion.
        """
        proj = project_obj.Project(self.ctx, id=1, external_id=self.project_id)
        self.mock_project_get.return_value = proj
        user = user_obj.User(self.ctx, id=1, external_id=self.user_id)
        self.mock_user_get.return_value = user
        self.mock_consumer_get.side_effect = exception.NotFound

        consumer_gen = None  # should be ignored
        util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.before_version)

        self.mock_project_create.assert_not_called()
        self.mock_user_create.assert_not_called()
        self.mock_consumer_create.assert_called_once()

    def test_existing_consumer_after_gen_matches_supplied_gen(self):
        """Tests that we require a consumer_generation after the
        appropriate microversion and that when the consumer already exists,
        then we ensure a matching generation is supplied
        """
        proj = project_obj.Project(self.ctx, id=1, external_id=self.project_id)
        self.mock_project_get.return_value = proj
        user = user_obj.User(self.ctx, id=1, external_id=self.user_id)
        self.mock_user_get.return_value = user
        consumer = consumer_obj.Consumer(
            self.ctx, id=1, project=proj, user=user, generation=2)
        self.mock_consumer_get.return_value = consumer

        consumer_gen = 2  # should NOT be ignored (and 2 is expected)
        util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.after_version)

        self.mock_project_create.assert_not_called()
        self.mock_user_create.assert_not_called()
        self.mock_consumer_create.assert_not_called()

    def test_existing_consumer_after_gen_fail(self):
        """Tests that we require a consumer_generation after the
        appropriate microversion and that when the consumer already exists,
        then we raise a 400 when there is a mismatch on the existing
        generation.
        """
        proj = project_obj.Project(self.ctx, id=1, external_id=self.project_id)
        self.mock_project_get.return_value = proj
        user = user_obj.User(self.ctx, id=1, external_id=self.user_id)
        self.mock_user_get.return_value = user
        consumer = consumer_obj.Consumer(
            self.ctx, id=1, project=proj, user=user, generation=42)
        self.mock_consumer_get.return_value = consumer

        consumer_gen = 2  # should NOT be ignored (and 2 is NOT expected)
        self.assertRaises(
            webob.exc.HTTPConflict,
            util.ensure_consumer,
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.after_version)

    def test_existing_consumer_different_consumer_type_supplied(self):
        """Tests that we update a consumer's type ID if the one supplied by the
        user is different than the one in the existing record.
        """
        proj = project_obj.Project(self.ctx, id=1, external_id=self.project_id)
        self.mock_project_get.return_value = proj
        user = user_obj.User(self.ctx, id=1, external_id=self.user_id)
        self.mock_user_get.return_value = user
        # Consumer currently has type ID = 1
        consumer = consumer_obj.Consumer(
            self.ctx, id=1, project=proj, user=user, generation=1,
            consumer_type_id=1)
        self.mock_consumer_get.return_value = consumer

        consumer_gen = 1
        consumer, created_new_consumer, request_attr = util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.cons_type_req_version)
        util.update_consumers([consumer], {consumer.uuid: request_attr})
        # Expect 1 call to update() to update to the supplied consumer type ID
        self.mock_consumer_update.assert_called_once_with()
        # Consumer should have the new consumer type from the cache
        self.assertEqual(
            self.ctx.ct_cache.id_from_string.return_value,
            consumer.consumer_type_id)

    def test_consumer_create_exists_different_consumer_type_supplied(self):
        """Tests that we update a consumer's type ID if the one supplied by a
        racing request is different than the one in the existing (recently
        created) record.
        """
        proj = project_obj.Project(self.ctx, id=1, external_id=self.project_id)
        self.mock_project_get.return_value = proj
        user = user_obj.User(self.ctx, id=1, external_id=self.user_id)
        self.mock_user_get.return_value = user
        # Request A recently created consumer has type ID = 1
        consumer = consumer_obj.Consumer(
            self.ctx, id=1, project=proj, user=user, generation=1,
            consumer_type_id=1, uuid=uuidsentinel.consumer)
        self.mock_consumer_get.return_value = consumer
        # Request B will encounter ConsumerExists as Request A just created it
        self.mock_consumer_create.side_effect = (
            exception.ConsumerExists(uuid=uuidsentinel.consumer))

        consumer_gen = 1
        consumer, created_new_consumer, request_attr = util.ensure_consumer(
            self.ctx, self.consumer_id, self.project_id, self.user_id,
            consumer_gen, 'TYPE', self.cons_type_req_version)
        util.update_consumers([consumer], {consumer.uuid: request_attr})
        # Expect 1 call to update() to update to the supplied consumer type ID
        self.mock_consumer_update.assert_called_once_with()
        # Consumer should have the new consumer type from the cache
        self.assertEqual(
            self.ctx.ct_cache.id_from_string.return_value,
            consumer.consumer_type_id)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

workarounds_group = cfg.OptGroup(
    'workarounds',
    title='Workaround Options',
    help="""
A collection of workarounds used to mitigate bugs or issues found under
certain conditions. These should only be enabled in exceptional circumstances.
All options are linked against bug IDs, where more information on the issue can
be found.
""")

workaround_opts = [
    cfg.BoolOpt(
        "optimize_for_wide_provider_trees",
        default=False,
        help="""
Enable optimization of allocation candidate generation for wide provider trees.

As reported in `bug #2126751`_ in the situation where many similar child
provider is defined under the same root provider, placement's allocation
candidate generation algorithm scales poorly. This config option enables
certain optimizations that help decrease the time it takes to generate the
GET /allocation_candidates response for queries requesting multiple resources
from those child providers.

For example if a compute has 8 or more child resource providers providing one
resource each (e.g. 8 individual PGPU) and a VM requests 8 or more such
resources each in independent request groups then without this optimization
enabled the GET /allocation_candidates query takes too long to compute and
the scheduling will fail.

Setting the ``[placement]max_allocation_candidates``
config option to a small number (e.g. 100) can help to a certain degree but
alone cannot solve the problem when the number of devices available or the
number of requested devices increases.

**When to enable:** If you have at least 8 child resource providers within a
tree providing inventory of the same resource class. And you are trying
to support VMs with more than 4 such resources.
E.g.:

* Nova's PCI in Placement feature is enabled and you have at least 8 PCI
  devices with the same product_id in a single compute and you are using
  flavors requesting more than 4 such devices.

* Nova's GPU support is enabled and you have at least 8 GPUs per compute node
  while requesting more than 4 per VM.

**When not to enable:** If you have a flat resource provider tree, i.e. all
resources reported on the root provider. Or if your flavors are not requesting
more than 4 PCI or GPU resources of the same type.

Related options:

* ``[placement]max_allocation_candidates``: If you need to enable the
  this optimization then you are also in a situation where you want to set
  ``max_allocation_candidates`` to a number not more than 1000.
* ``[placement]allocation_candidates_generation_strategy``: If you use
  ``max_allocation_candidates`` then it is suggested to configure
  ``allocation_candidates_generation_strategy`` to ``breadth-first`` which will
  return candidates balanced across available compute nodes.

.. _bug #2126751: https://bugs.launchpad.net/placement/+bug/2126751
"""),
]


def register_opts(conf):
    conf.register_group(workarounds_group)
    conf.register_opts(workaround_opts, group=workarounds_group)


def list_opts():
    return {workarounds_group: workaround_opts}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Placement API handlers for resource classes."""

from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import timeutils
import webob

from placement import exception
from placement import microversion
from placement.objects import resource_class as rc_obj
from placement.policies import resource_class as policies
from placement.schemas import resource_class as schema
from placement import util
from placement import wsgi_wrapper


def _serialize_links(environ, rc):
    url = util.resource_class_url(environ, rc)
    links = [{'rel': 'self', 'href': url}]
    return links


def _serialize_resource_class(environ, rc):
    data = {
        'name': rc.name,
        'links': _serialize_links(environ, rc)
    }
    return data


def _serialize_resource_classes(environ, rcs, want_version):
    output = []
    last_modified = None
    get_last_modified = want_version.matches((1, 15))
    for rc in rcs:
        if get_last_modified:
            last_modified = util.pick_last_modified(last_modified, rc)
        data = _serialize_resource_class(environ, rc)
        output.append(data)
    last_modified = last_modified or timeutils.utcnow(with_timezone=True)
    return ({"resource_classes": output}, last_modified)


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.2')
@util.require_content('application/json')
def create_resource_class(req):
    """POST to create a resource class.

    On success return a 201 response with an empty body and a location
    header pointing to the newly created resource class.
    """
    context = req.environ['placement.context']
    context.can(policies.CREATE)
    data = util.extract_json(req.body, schema.POST_RC_SCHEMA_V1_2)

    try:
        rc = rc_obj.ResourceClass(context, name=data['name'])
        rc.create()
    except exception.ResourceClassExists:
        raise webob.exc.HTTPConflict(
            'Conflicting resource class already exists: %(name)s' %
            {'name': data['name']})
    except exception.MaxDBRetriesExceeded:
        raise webob.exc.HTTPConflict(
            'Max retries of DB transaction exceeded attempting '
            'to create resource class: %(name)s, please '
            'try again.' %
            {'name': data['name']})

    req.response.location = util.resource_class_url(req.environ, rc)
    req.response.status = 201
    req.response.content_type = None
    return req.response


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.2')
def delete_resource_class(req):
    """DELETE to destroy a single resource class.

    On success return a 204 and an empty body.
    """
    name = util.wsgi_path_item(req.environ, 'name')
    context = req.environ['placement.context']
    context.can(policies.DELETE)
    # The containing application will catch a not found here.
    rc = rc_obj.ResourceClass.get_by_name(context, name)
    try:
        rc.destroy()
    except exception.ResourceClassCannotDeleteStandard as exc:
        raise webob.exc.HTTPBadRequest(
            'Error in delete resource class: %(error)s' % {'error': exc})
    except exception.ResourceClassInUse as exc:
        raise webob.exc.HTTPConflict(
            'Error in delete resource class: %(error)s' % {'error': exc})
    req.response.status = 204
    req.response.content_type = None
    return req.response


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.2')
@util.check_accept('application/json')
def get_resource_class(req):
    """Get a single resource class.

    On success return a 200 with an application/json body representing
    the resource class.
    """
    name = util.wsgi_path_item(req.environ, 'name')
    context = req.environ['placement.context']
    context.can(policies.SHOW)
    want_version = req.environ[microversion.MICROVERSION_ENVIRON]
    # The containing application will catch a not found here.
    rc = rc_obj.ResourceClass.get_by_name(context, name)

    req.response.body = encodeutils.to_utf8(jsonutils.dumps(
        _serialize_resource_class(req.environ, rc))
    )
    req.response.content_type = 'application/json'
    if want_version.matches((1, 15)):
        req.response.cache_control = 'no-cache'
        # Non-custom resource classes will return None from pick_last_modified,
        # so the 'or' causes utcnow to be used.
        last_modified = util.pick_last_modified(None, rc) or timeutils.utcnow(
            with_timezone=True)
        req.response.last_modified = last_modified
    return req.response


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.2')
@util.check_accept('application/json')
def list_resource_classes(req):
    """GET a list of resource classes.

    On success return a 200 and an application/json body representing
    a collection of resource classes.
    """
    context = req.environ['placement.context']
    context.can(policies.LIST)
    want_version = req.environ[microversion.MICROVERSION_ENVIRON]
    rcs = rc_obj.get_all(context)

    response = req.response
    output, last_modified = _serialize_resource_classes(
        req.environ, rcs, want_version)
    response.body = encodeutils.to_utf8(jsonutils.dumps(output))
    response.content_type = 'application/json'
    if want_version.matches((1, 15)):
        response.last_modified = last_modified
        response.cache_control = 'no-cache'
    return response


@wsgi_wrapper.PlacementWsgify
@microversion.version_handler('1.2', '1.6')
@util.require_content('application/json')
def update_resource_class(req):
    """PUT to update a single resource class.

    On success return a 200 response with a representation of the updated
    resource class.
    """
    name = util.wsgi_path_item(req.environ, 'name')
    context = req.environ['placement.context']
    context.can(policies.UPDATE)

    data = util.extract_json(req.body, schema.PUT_RC_SCHEMA_V1_2)

    # The containing application will catch a not found here.
    rc = rc_obj.ResourceClass.get_by_name(context, name)

    rc.name = data['name']

    try:
        rc.save()
    except exception.ResourceClassExists:
        raise webob.exc.HTTPConflict(
            'Resource class already exists: %(name)s' %
            {'name': rc.name})
    except exception.ResourceClassCannotUpdateStandard:
        raise webob.exc.HTTPBadRequest(
            'Cannot update standard resource class %(rp_name)s' %
            {'rp_name': name})

    req.response.body = encodeutils.to_utf8(jsonutils.dumps(
        _serialize_resource_class(req.environ, rc))
    )
    req.response.status = 200
    req.response.content_type = 'application/json'
    return req.response


@wsgi_wrapper.PlacementWsgify  # noqa
@microversion.version_handler('1.7')
def update_resource_class(req):  # noqa
    """PUT to create or validate the existence of single resource class.

    On a successful create return 201. Return 204 if the class already
    exists. If the resource class is not a custom resource class, return
    a 400. 409 might be a better choice, but 400 aligns with previous code.
    """
    name = util.wsgi_path_item(req.environ, 'name')
    context = req.environ['placement.context']
    context.can(policies.UPDATE)

    # Use JSON validation to validation resource class name.
    util.extract_json('{"name": "%s"}' % name, schema.PUT_RC_SCHEMA_V1_2)

    status = 204
    try:
        rc = rc_obj.ResourceClass.get_by_name(context, name)
    except exception.NotFound:
        try:
            rc = rc_obj.ResourceClass(context, name=name)
            rc.create()
            status = 201
        # We will not see ResourceClassCannotUpdateStandard because
        # that was already caught when validating the {name}.
        except exception.ResourceClassExists:
            # Someone just now created the class, so stick with 204
            pass

    req.response.status = status
    req.response.content_type = None
    req.response.location = util.resource_class_url(req.environ, rc)
    return req.response



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Placement API schemas for getting allocation candidates."""

import copy

from placement.schemas import common


# Represents the allowed query string parameters to the GET
# /allocation_candidates API call
GET_SCHEMA_1_10 = {
    "type": "object",
    "properties": {
        "resources": {
            "type": "string"
        },
    },
    "required": [
        "resources",
    ],
    "additionalProperties": False,
}


# Add limit query parameter.
GET_SCHEMA_1_16 = copy.deepcopy(GET_SCHEMA_1_10)
GET_SCHEMA_1_16['properties']['limit'] = {
    # A query parameter is always a string in webOb, but
    # we'll handle integer here as well.
    "type": ["integer", "string"],
    "pattern": "^[1-9][0-9]*$",
    "minimum": 1,
    "minLength": 1
}

# Add required parameter.
GET_SCHEMA_1_17 = copy.deepcopy(GET_SCHEMA_1_16)
GET_SCHEMA_1_17['properties']['required'] = {
    "type": ["string"]
}

# Add member_of parameter.
GET_SCHEMA_1_21 = copy.deepcopy(GET_SCHEMA_1_17)
GET_SCHEMA_1_21['properties']['member_of'] = {
    "type": ["string"]
}

GET_SCHEMA_1_25 = copy.deepcopy(GET_SCHEMA_1_21)
# We're going to *replace* 'resources', 'required', and 'member_of'.
del GET_SCHEMA_1_25["properties"]["resources"]
del GET_SCHEMA_1_25["required"]
del GET_SCHEMA_1_25["properties"]["required"]
del GET_SCHEMA_1_25["properties"]["member_of"]
# Pattern property key format for a numbered or un-numbered grouping
_GROUP_PAT_FMT = "^%s(" + common.GROUP_PAT + ")?$"
GET_SCHEMA_1_25["patternProperties"] = {
    _GROUP_PAT_FMT % "resources": {
        "type": "string",
    },
    _GROUP_PAT_FMT % "required": {
        "type": "string",
    },
    _GROUP_PAT_FMT % "member_of": {
        "type": "string",
    },
}
GET_SCHEMA_1_25["properties"]["group_policy"] = {
    "type": "string",
    "enum": ["none", "isolate"],
}

# Add in_tree parameter.
GET_SCHEMA_1_31 = copy.deepcopy(GET_SCHEMA_1_25)
GET_SCHEMA_1_31["patternProperties"][_GROUP_PAT_FMT % "in_tree"] = {
    "type": "string"}

# Microversion 1.33 allows more complex resource group suffixes.
GET_SCHEMA_1_33 = copy.deepcopy(GET_SCHEMA_1_31)
_GROUP_PAT_FMT_1_33 = "^%s(" + common.GROUP_PAT_1_33 + ")?$"
GET_SCHEMA_1_33["patternProperties"] = {
    _GROUP_PAT_FMT_1_33 % group_type: {"type": "string"}
    for group_type in ('resources', 'required', 'member_of', 'in_tree')}

# Microversion 1.35 supports root_required.
GET_SCHEMA_1_35 = copy.deepcopy(GET_SCHEMA_1_33)
GET_SCHEMA_1_35["properties"]['root_required'] = {
    "type": ["string"]
}

# Microversion 1.36 supports same_subtree.
GET_SCHEMA_1_36 = copy.deepcopy(GET_SCHEMA_1_35)
GET_SCHEMA_1_36["properties"]['same_subtree'] = {
    "type": ["string"]
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Handlers for placement API.

Individual handlers are associated with URL paths in the
ROUTE_DECLARATIONS dictionary. At the top level each key is a Routes
compliant path. The value of that key is a dictionary mapping
individual HTTP request methods to a Python function representing a
simple WSGI application for satisfying that request.

The ``make_map`` method processes ROUTE_DECLARATIONS to create a
Routes.Mapper, including automatic handlers to respond with a
405 when a request is made against a valid URL with an invalid
method.
"""

import routes
import webob

from oslo_log import log as logging

from placement import exception
from placement.handlers import aggregate
from placement.handlers import allocation
from placement.handlers import allocation_candidate
from placement.handlers import inventory
from placement.handlers import reshaper
from placement.handlers import resource_class
from placement.handlers import resource_provider
from placement.handlers import root
from placement.handlers import trait
from placement.handlers import usage
from placement import util

LOG = logging.getLogger(__name__)

# URLs and Handlers
# NOTE(cdent): When adding URLs here, do not use regex patterns in
# the path parameters (e.g. {uuid:[0-9a-zA-Z-]+}) as that will lead
# to 404s that are controlled outside of the individual resources
# and thus do not include specific information on the why of the 404.
ROUTE_DECLARATIONS = {
    '/': {
        'GET': root.home,
    },
    # NOTE(cdent): This allows '/placement/' and '/placement' to
    # both work as the root of the service, which we probably want
    # for those situations where the service is mounted under a
    # prefix (as it is in devstack). While weird, an empty string is
    # a legit key in a dictionary and matches as desired in Routes.
    '': {
        'GET': root.home,
    },
    '/resource_classes': {
        'GET': resource_class.list_resource_classes,
        'POST': resource_class.create_resource_class
    },
    '/resource_classes/{name}': {
        'GET': resource_class.get_resource_class,
        'PUT': resource_class.update_resource_class,
        'DELETE': resource_class.delete_resource_class,
    },
    '/resource_providers': {
        'GET': resource_provider.list_resource_providers,
        'POST': resource_provider.create_resource_provider
    },
    '/resource_providers/{uuid}': {
        'GET': resource_provider.get_resource_provider,
        'DELETE': resource_provider.delete_resource_provider,
        'PUT': resource_provider.update_resource_provider
    },
    '/resource_providers/{uuid}/inventories': {
        'GET': inventory.get_inventories,
        'POST': inventory.create_inventory,
        'PUT': inventory.set_inventories,
        'DELETE': inventory.delete_inventories
    },
    '/resource_providers/{uuid}/inventories/{resource_class}': {
        'GET': inventory.get_inventory,
        'PUT': inventory.update_inventory,
        'DELETE': inventory.delete_inventory
    },
    '/resource_providers/{uuid}/usages': {
        'GET': usage.list_usages
    },
    '/resource_providers/{uuid}/aggregates': {
        'GET': aggregate.get_aggregates,
        'PUT': aggregate.set_aggregates
    },
    '/resource_providers/{uuid}/allocations': {
        'GET': allocation.list_for_resource_provider,
    },
    '/allocations': {
        'POST': allocation.set_allocations,
    },
    '/allocations/{consumer_uuid}': {
        'GET': allocation.list_for_consumer,
        'PUT': allocation.set_allocations_for_consumer,
        'DELETE': allocation.delete_allocations,
    },
    '/allocation_candidates': {
        'GET': allocation_candidate.list_allocation_candidates,
    },
    '/traits': {
        'GET': trait.list_traits,
    },
    '/traits/{name}': {
        'GET': trait.get_trait,
        'PUT': trait.put_trait,
        'DELETE': trait.delete_trait,
    },
    '/resource_providers/{uuid}/traits': {
        'GET': trait.list_traits_for_resource_provider,
        'PUT': trait.update_traits_for_resource_provider,
        'DELETE': trait.delete_traits_for_resource_provider
    },
    '/usages': {
        'GET': usage.get_total_usages,
    },
    '/reshaper': {
        'POST': reshaper.reshape,
    },
}


def dispatch(environ, start_response, mapper):
    """Find a matching route for the current request.

    If no match is found, raise a 404 response.
    If there is a matching route, but no matching handler
    for the given method, raise a 405.
    """
    result = mapper.match(environ=environ)
    if result is None:
        raise webob.exc.HTTPNotFound(
            json_formatter=util.json_error_formatter)
    # We can't reach this code without action being present.
    handler = result.pop('action')
    environ['wsgiorg.routing_args'] = ((), result)
    return handler(environ, start_response)


def handle_405(environ, start_response):
    """Return a 405 response when method is not allowed.

    If _methods are in routing_args, send an allow header listing
    the methods that are possible on the provided URL.
    """
    _methods = util.wsgi_path_item(environ, '_methods')
    headers = {}
    if _methods:
        # Ensure allow header is a python 2 or 3 native string (thus
        # not unicode in python 2 but stay a string in python 3)
        # In the process done by Routes to save the allowed methods
        # to its routing table they become unicode in py2.
        headers['allow'] = str(_methods)
    # Use Exception class as WSGI Application. We don't want to raise here.
    response = webob.exc.HTTPMethodNotAllowed(
        'The method specified is not allowed for this resource.',
        headers=headers, json_formatter=util.json_error_formatter)
    return response(environ, start_response)


def make_map(declarations):
    """Process route declarations to create a Route Mapper."""
    mapper = routes.Mapper()
    for route, targets in declarations.items():
        allowed_methods = []
        for method in targets:
            mapper.connect(route, action=targets[method],
                           conditions=dict(method=[method]))
            allowed_methods.append(method)
        allowed_methods = ', '.join(allowed_methods)
        mapper.connect(route, action=handle_405, _methods=allowed_methods)
    return mapper


class PlacementHandler(object):
    """Serve Placement API.

    Dispatch to handlers defined in ROUTE_DECLARATIONS.
    """

    def __init__(self, **local_config):
        self._map = make_map(ROUTE_DECLARATIONS)
        self.config = local_config['config']

    def __call__(self, environ, start_response):
        # set a reference to the oslo.config ConfigOpts on the RequestContext
        context = environ['placement.context']
        context.config = self.config
        # Check that an incoming request with a content-length header
        # that is an integer > 0 and not empty, also has a content-type
        # header that is not empty. If not raise a 400.
        clen = environ.get('CONTENT_LENGTH')
        try:
            if clen and (int(clen) > 0) and not environ.get('CONTENT_TYPE'):
                raise webob.exc.HTTPBadRequest(
                    'content-type header required when content-length > 0',
                    json_formatter=util.json_error_formatter)
        except ValueError:
            raise webob.exc.HTTPBadRequest(
                'content-length header must be an integer',
                json_formatter=util.json_error_formatter)
        try:
            return dispatch(environ, start_response, self._map)
        # Trap the NotFound exceptions raised by the objects used
        # with the API and transform them into webob.exc.HTTPNotFound.
        except exception.NotFound as exc:
            raise webob.exc.HTTPNotFound(
                exc, json_formatter=util.json_error_formatter)
        except exception.PolicyNotAuthorized as exc:
            raise webob.exc.HTTPForbidden(
                exc.format_message(),
                json_formatter=util.json_error_formatter)
        # Remaining uncaught exceptions will rise first to the Microversion
        # middleware, where any WebOb generated exceptions will be caught and
        # transformed into legit HTTP error responses (with microversion
        # headers added), and then to the FaultWrapper middleware which will
        # catch anything else and transform them into 500 responses.
        # NOTE(cdent): There should be very few uncaught exceptions which are
        # not WebOb exceptions at this stage as the handlers are contained by
        # the wsgify decorator which will transform those exceptions to
        # responses itself.



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from placement.objects import rp_candidates


class TestRPCandidateList(testtools.TestCase):
    def setUp(self):
        super(TestRPCandidateList, self).setUp()
        self.rp_candidates = rp_candidates.RPCandidateList()
        self.rps_rc1 = set([
            ('rp1', 'root1'), ('rp2', 'root1'), ('ss1', 'root1'),
            ('rp3', 'root'), ('ss1', 'root')])
        self.rp_candidates.add_rps(self.rps_rc1, 'rc_1')

    def test_property(self):
        expected_rpsinfo = set([('rp1', 'root1', 'rc_1'),
                                ('rp2', 'root1', 'rc_1'),
                                ('ss1', 'root1', 'rc_1'),
                                ('rp3', 'root', 'rc_1'),
                                ('ss1', 'root', 'rc_1')])
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)

        expected_rps = set(['rp1', 'rp2', 'rp3', 'ss1'])
        expected_trees = set(['root1', 'root'])
        expected_allrps = expected_rps | expected_trees

        self.assertEqual(expected_rps, self.rp_candidates.rps)
        self.assertEqual(expected_trees, self.rp_candidates.trees)
        self.assertEqual(expected_allrps, self.rp_candidates.all_rps)

    def test_filter_by_tree(self):
        self.rp_candidates.filter_by_tree(set(['root1']))
        expected_rpsinfo = set([('rp1', 'root1', 'rc_1'),
                                ('rp2', 'root1', 'rc_1'),
                                ('ss1', 'root1', 'rc_1')])
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)

    def test_filter_by_rp(self):
        self.rp_candidates.filter_by_rp(set([('ss1', 'root1')]))
        expected_rpsinfo = set([('ss1', 'root1', 'rc_1')])
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)

    def test_filter_by_rp_or_tree(self):
        self.rp_candidates.filter_by_rp_or_tree(set(['ss1', 'root1']))
        # we get 'ss1' and rps under 'root1'
        expected_rpsinfo = set([('ss1', 'root', 'rc_1'),
                                ('ss1', 'root1', 'rc_1'),
                                ('rp1', 'root1', 'rc_1'),
                                ('rp2', 'root1', 'rc_1')])
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)

    def test_merge_common_trees(self):
        merge_candidates = rp_candidates.RPCandidateList()
        rps_rc2 = set([('rp1', 'root2'), ('rp4', 'root2'), ('ss1', 'root2'),
                       ('rp5', 'root'), ('ss1', 'root')])
        merge_candidates.add_rps(rps_rc2, 'rc_2')

        self.rp_candidates.merge_common_trees(merge_candidates)
        # we get only rps under 'root' since it's only the common tree
        expected_rpsinfo = set([('rp3', 'root', 'rc_1'),
                                ('rp5', 'root', 'rc_2'),
                                ('ss1', 'root', 'rc_1'),
                                ('ss1', 'root', 'rc_2')])
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)

        # make sure merging empty candidates doesn't change anything
        empty_candidates = rp_candidates.RPCandidateList()
        self.rp_candidates.merge_common_trees(empty_candidates)
        self.assertEqual(expected_rpsinfo, self.rp_candidates.rps_info)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_utils.fixture import uuidsentinel as uuids

from placement import exception
from placement.objects import allocation as alloc_obj
from placement.objects import consumer as consumer_obj
from placement.objects import inventory as inv_obj
from placement.objects import reshaper
from placement.objects import resource_provider as rp_obj
from placement.tests.functional.db import test_base as tb


def alloc_for_rc(alloc_list, rc):
    for alloc in alloc_list:
        if alloc.resource_class == rc:
            return alloc


class ReshapeTestCase(tb.PlacementDbBaseTestCase):
    """Test 'replace the world' reshape transaction."""

    def test_reshape(self):
        """We set up the following scenario:

        BEFORE: single compute node setup

          A single compute node with:
            - VCPU, MEMORY_MB, DISK_GB inventory
            - Two instances consuming CPU, RAM and DISK from that compute node

        AFTER: hierarchical + shared storage setup

          A compute node parent provider with:
            - MEMORY_MB
          Two NUMA node child providers containing:
            - VCPU
          Shared storage provider with:
            - DISK_GB
          Both instances have their resources split among the providers and
          shared storage accordingly
        """
        # First create our consumers
        i1_uuid = uuids.instance1
        i1_consumer = consumer_obj.Consumer(
            self.ctx, uuid=i1_uuid, user=self.user_obj,
            project=self.project_obj)
        i1_consumer.create()

        i2_uuid = uuids.instance2
        i2_consumer = consumer_obj.Consumer(
            self.ctx, uuid=i2_uuid, user=self.user_obj,
            project=self.project_obj)
        i2_consumer.create()

        cn1 = self._create_provider('cn1')
        tb.add_inventory(cn1, 'VCPU', 16)
        tb.add_inventory(cn1, 'MEMORY_MB', 32768)
        tb.add_inventory(cn1, 'DISK_GB', 1000)

        # Allocate both instances against the single compute node
        for consumer in (i1_consumer, i2_consumer):
            allocs = [
                alloc_obj.Allocation(
                    resource_provider=cn1,
                    resource_class='VCPU', consumer=consumer, used=2),
                alloc_obj.Allocation(
                    resource_provider=cn1,
                    resource_class='MEMORY_MB', consumer=consumer, used=1024),
                alloc_obj.Allocation(
                    resource_provider=cn1,
                    resource_class='DISK_GB', consumer=consumer, used=100),
            ]
            alloc_obj.replace_all(self.ctx, allocs)

        # Verify we have the allocations we expect for the BEFORE scenario
        before_allocs_i1 = alloc_obj.get_all_by_consumer_id(self.ctx, i1_uuid)
        self.assertEqual(3, len(before_allocs_i1))
        self.assertEqual(cn1.uuid, before_allocs_i1[0].resource_provider.uuid)
        before_allocs_i2 = alloc_obj.get_all_by_consumer_id(self.ctx, i2_uuid)
        self.assertEqual(3, len(before_allocs_i2))
        self.assertEqual(cn1.uuid, before_allocs_i2[2].resource_provider.uuid)

        # Before we issue the actual reshape() call, we need to first create
        # the child providers and sharing storage provider. These are actions
        # that the virt driver or external agent is responsible for performing
        # *before* attempting any reshape activity.
        cn1_numa0 = self._create_provider('cn1_numa0', parent=cn1.uuid)
        cn1_numa1 = self._create_provider('cn1_numa1', parent=cn1.uuid)
        ss = self._create_provider('ss')

        # OK, now emulate the call to POST /reshaper that will be triggered by
        # a virt driver wanting to replace the world and change its modeling
        # from a single provider to a nested provider tree along with a sharing
        # storage provider.
        after_inventories = {
            # cn1 keeps the RAM only
            cn1: [
                inv_obj.Inventory(
                    resource_provider=cn1,
                    resource_class='MEMORY_MB', total=32768, reserved=0,
                    max_unit=32768, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            # each NUMA node gets half of the CPUs
            cn1_numa0: [
                inv_obj.Inventory(
                    resource_provider=cn1_numa0,
                    resource_class='VCPU', total=8, reserved=0,
                    max_unit=8, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            cn1_numa1: [
                inv_obj.Inventory(
                    resource_provider=cn1_numa1,
                    resource_class='VCPU', total=8, reserved=0,
                    max_unit=8, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            # The sharing provider gets a bunch of disk
            ss: [
                inv_obj.Inventory(
                    resource_provider=ss,
                    resource_class='DISK_GB', total=100000, reserved=0,
                    max_unit=1000, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
        }
        # We do a fetch from the DB for each instance to get its latest
        # generation. This would be done by the resource tracker or scheduler
        # report client before issuing the call to reshape() because the
        # consumers representing the two instances above will have had their
        # generations incremented in the original call to PUT
        # /allocations/{consumer_uuid}
        i1_consumer = consumer_obj.Consumer.get_by_uuid(self.ctx, i1_uuid)
        i2_consumer = consumer_obj.Consumer.get_by_uuid(self.ctx, i2_uuid)
        after_allocs = [
            # instance1 gets VCPU from NUMA0, MEMORY_MB from cn1 and DISK_GB
            # from the sharing storage provider
            alloc_obj.Allocation(
                resource_provider=cn1_numa0, resource_class='VCPU',
                consumer=i1_consumer, used=2),
            alloc_obj.Allocation(
                resource_provider=cn1, resource_class='MEMORY_MB',
                consumer=i1_consumer, used=1024),
            alloc_obj.Allocation(
                resource_provider=ss, resource_class='DISK_GB',
                consumer=i1_consumer, used=100),
            # instance2 gets VCPU from NUMA1, MEMORY_MB from cn1 and DISK_GB
            # from the sharing storage provider
            alloc_obj.Allocation(
                resource_provider=cn1_numa1, resource_class='VCPU',
                consumer=i2_consumer, used=2),
            alloc_obj.Allocation(
                resource_provider=cn1, resource_class='MEMORY_MB',
                consumer=i2_consumer, used=1024),
            alloc_obj.Allocation(
                resource_provider=ss, resource_class='DISK_GB',
                consumer=i2_consumer, used=100),
        ]
        reshaper.reshape(self.ctx, after_inventories, after_allocs)

        # Verify that the inventories have been moved to the appropriate
        # providers in the AFTER scenario

        # The root compute node should only have MEMORY_MB, nothing else
        cn1_inv = inv_obj.get_all_by_resource_provider(self.ctx, cn1)
        self.assertEqual(1, len(cn1_inv))
        self.assertEqual('MEMORY_MB', cn1_inv[0].resource_class)
        self.assertEqual(32768, cn1_inv[0].total)
        # Each NUMA node should only have half the original VCPU, nothing else
        numa0_inv = inv_obj.get_all_by_resource_provider(self.ctx, cn1_numa0)
        self.assertEqual(1, len(numa0_inv))
        self.assertEqual('VCPU', numa0_inv[0].resource_class)
        self.assertEqual(8, numa0_inv[0].total)
        numa1_inv = inv_obj.get_all_by_resource_provider(self.ctx, cn1_numa1)
        self.assertEqual(1, len(numa1_inv))
        self.assertEqual('VCPU', numa1_inv[0].resource_class)
        self.assertEqual(8, numa1_inv[0].total)
        # The sharing storage provider should only have DISK_GB, nothing else
        ss_inv = inv_obj.get_all_by_resource_provider(self.ctx, ss)
        self.assertEqual(1, len(ss_inv))
        self.assertEqual('DISK_GB', ss_inv[0].resource_class)
        self.assertEqual(100000, ss_inv[0].total)

        # Verify we have the allocations we expect for the AFTER scenario
        after_allocs_i1 = alloc_obj.get_all_by_consumer_id(self.ctx, i1_uuid)
        self.assertEqual(3, len(after_allocs_i1))
        # Our VCPU allocation should be in the NUMA0 node
        vcpu_alloc = alloc_for_rc(after_allocs_i1, 'VCPU')
        self.assertIsNotNone(vcpu_alloc)
        self.assertEqual(cn1_numa0.uuid, vcpu_alloc.resource_provider.uuid)
        # Our DISK_GB allocation should be in the sharing provider
        disk_alloc = alloc_for_rc(after_allocs_i1, 'DISK_GB')
        self.assertIsNotNone(disk_alloc)
        self.assertEqual(ss.uuid, disk_alloc.resource_provider.uuid)
        # And our MEMORY_MB should remain on the root compute node
        ram_alloc = alloc_for_rc(after_allocs_i1, 'MEMORY_MB')
        self.assertIsNotNone(ram_alloc)
        self.assertEqual(cn1.uuid, ram_alloc.resource_provider.uuid)

        after_allocs_i2 = alloc_obj.get_all_by_consumer_id(self.ctx, i2_uuid)
        self.assertEqual(3, len(after_allocs_i2))
        # Our VCPU allocation should be in the NUMA1 node
        vcpu_alloc = alloc_for_rc(after_allocs_i2, 'VCPU')
        self.assertIsNotNone(vcpu_alloc)
        self.assertEqual(cn1_numa1.uuid, vcpu_alloc.resource_provider.uuid)
        # Our DISK_GB allocation should be in the sharing provider
        disk_alloc = alloc_for_rc(after_allocs_i2, 'DISK_GB')
        self.assertIsNotNone(disk_alloc)
        self.assertEqual(ss.uuid, disk_alloc.resource_provider.uuid)
        # And our MEMORY_MB should remain on the root compute node
        ram_alloc = alloc_for_rc(after_allocs_i2, 'MEMORY_MB')
        self.assertIsNotNone(ram_alloc)
        self.assertEqual(cn1.uuid, ram_alloc.resource_provider.uuid)

    def test_reshape_concurrent_inventory_update(self):
        """Valid failure scenario for reshape(). We test a situation where the
        virt driver has constructed it's "after inventories and allocations"
        and sent those to the POST /reshape endpoint. The reshape POST handler
        does a quick check of the resource provider generations sent in the
        payload and they all check out.

        However, right before the call to resource_provider.reshape(), another
        thread legitimately changes the inventory of one of the providers
        involved in the reshape transaction. We should get a
        ConcurrentUpdateDetected in this case.
        """
        # First create our consumers
        i1_uuid = uuids.instance1
        i1_consumer = consumer_obj.Consumer(
            self.ctx, uuid=i1_uuid, user=self.user_obj,
            project=self.project_obj)
        i1_consumer.create()

        # then all our original providers
        cn1 = self._create_provider('cn1')
        tb.add_inventory(cn1, 'VCPU', 16)
        tb.add_inventory(cn1, 'MEMORY_MB', 32768)
        tb.add_inventory(cn1, 'DISK_GB', 1000)

        # Allocate an instance on our compute node
        allocs = [
            alloc_obj.Allocation(
                resource_provider=cn1,
                resource_class='VCPU', consumer=i1_consumer, used=2),
            alloc_obj.Allocation(
                resource_provider=cn1,
                resource_class='MEMORY_MB', consumer=i1_consumer, used=1024),
            alloc_obj.Allocation(
                resource_provider=cn1,
                resource_class='DISK_GB', consumer=i1_consumer, used=100),
        ]
        alloc_obj.replace_all(self.ctx, allocs)

        # Before we issue the actual reshape() call, we need to first create
        # the child providers and sharing storage provider. These are actions
        # that the virt driver or external agent is responsible for performing
        # *before* attempting any reshape activity.
        cn1_numa0 = self._create_provider('cn1_numa0', parent=cn1.uuid)
        cn1_numa1 = self._create_provider('cn1_numa1', parent=cn1.uuid)
        ss = self._create_provider('ss')

        # OK, now emulate the call to POST /reshaper that will be triggered by
        # a virt driver wanting to replace the world and change its modeling
        # from a single provider to a nested provider tree along with a sharing
        # storage provider.
        after_inventories = {
            # cn1 keeps the RAM only
            cn1: [
                inv_obj.Inventory(
                    resource_provider=cn1,
                    resource_class='MEMORY_MB', total=32768, reserved=0,
                    max_unit=32768, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            # each NUMA node gets half of the CPUs
            cn1_numa0: [
                inv_obj.Inventory(
                    resource_provider=cn1_numa0,
                    resource_class='VCPU', total=8, reserved=0,
                    max_unit=8, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            cn1_numa1: [
                inv_obj.Inventory(
                    resource_provider=cn1_numa1,
                    resource_class='VCPU', total=8, reserved=0,
                    max_unit=8, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
            # The sharing provider gets a bunch of disk
            ss: [
                inv_obj.Inventory(
                    resource_provider=ss,
                    resource_class='DISK_GB', total=100000, reserved=0,
                    max_unit=1000, min_unit=1, step_size=1,
                    allocation_ratio=1.0),
            ],
        }
        # We do a fetch from the DB for each instance to get its latest
        # generation. This would be done by the resource tracker or scheduler
        # report client before issuing the call to reshape() because the
        # consumers representing the two instances above will have had their
        # generations incremented in the original call to PUT
        # /allocations/{consumer_uuid}
        i1_consumer = consumer_obj.Consumer.get_by_uuid(self.ctx, i1_uuid)
        after_allocs = [
            # instance1 gets VCPU from NUMA0, MEMORY_MB from cn1 and DISK_GB
            # from the sharing storage provider
            alloc_obj.Allocation(
                resource_provider=cn1_numa0, resource_class='VCPU',
                consumer=i1_consumer, used=2),
            alloc_obj.Allocation(
                resource_provider=cn1, resource_class='MEMORY_MB',
                consumer=i1_consumer, used=1024),
            alloc_obj.Allocation(
                resource_provider=ss, resource_class='DISK_GB',
                consumer=i1_consumer, used=100),
        ]

        # OK, now before we call reshape(), here we emulate another thread
        # changing the inventory for the sharing storage provider in between
        # the time in the REST handler when the sharing storage provider's
        # generation was validated and the actual call to reshape()
        ss_threadB = rp_obj.ResourceProvider.get_by_uuid(self.ctx, ss.uuid)
        # Reduce the amount of storage to 2000, from 100000.
        new_ss_inv = [
            inv_obj.Inventory(
                resource_provider=ss_threadB, resource_class='DISK_GB',
                total=2000, reserved=0, max_unit=1000, min_unit=1, step_size=1,
                allocation_ratio=1.0)]
        ss_threadB.set_inventory(new_ss_inv)
        # Double check our storage provider's generation is now greater than
        # the original storage provider record being sent to reshape()
        self.assertGreater(ss_threadB.generation, ss.generation)

        # And we should legitimately get a failure now to reshape() due to
        # another thread updating one of the involved provider's generations
        self.assertRaises(
            exception.ConcurrentUpdateDetected,
            reshaper.reshape, self.ctx, after_inventories, after_allocs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os_resource_classes
import os_traits

from placement import direct
from placement.tests.functional import base


class TestLibSync(base.TestCase):
    """Test that traits and resource classes are synced from os-traits and
    os-resource-classes libs to the DB at service startup.
    """

    def setUp(self):
        super().setUp()
        self.headers = {
            'x-auth-token': 'admin',
            'content-type': 'application/json',
            'OpenStack-API-Version': 'placement latest',
        }

    def test_traits_sync(self):
        with direct.PlacementDirect(self.conf_fixture.conf) as client:
            resp = client.get('/traits', headers=self.headers)
            self.assertCountEqual(
                os_traits.get_traits(),
                resp.json()['traits'],
            )

    def test_resource_classes_sync(self):
        with direct.PlacementDirect(self.conf_fixture.conf) as client:
            resp = client.get('/resource_classes', headers=self.headers)
            self.assertCountEqual(
                os_resource_classes.STANDARDS,
                [rc['name'] for rc in resp.json()['resource_classes']],
                resp.json(),
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Block on null consumer

Revision ID: b5c396305c25
Revises: 611cd6dffd7b
Create Date: 2019-06-11 16:30:04.114287

"""
from alembic import context
import sqlalchemy as sa
from sqlalchemy import func as sqlfunc


# revision identifiers, used by Alembic.
revision = 'b5c396305c25'
down_revision = '611cd6dffd7b'
branch_labels = None
depends_on = None


def upgrade():
    connection = context.get_bind()

    meta = sa.MetaData()
    meta.reflect(bind=connection)
    consumers = sa.Table('consumers', meta, autoload_with=connection)
    allocations = sa.Table('allocations', meta, autoload_with=connection)

    alloc_to_consumer = sa.outerjoin(
        allocations, consumers,
        allocations.c.consumer_id == consumers.c.uuid,
    )
    sel = sa.select(sqlfunc.count())
    sel = sel.select_from(alloc_to_consumer)
    sel = sel.where(consumers.c.id.is_(None))

    if connection.scalar(sel):
        raise Exception('There is at least one allocation record which is '
                        'missing a consumer record. Run the "placement-manage '
                        'db online_data_migrations" command.')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os_traits

from placement import exception
from placement.objects import trait as trait_obj
from placement.tests.functional.db import test_base as tb


class TraitTestCase(tb.PlacementDbBaseTestCase):

    def test_provider_traits_empty_param(self):
        self.assertRaises(ValueError, trait_obj.get_traits_by_provider_tree,
                          self.ctx, [])

    def test_trait_ids_from_names_empty_param(self):
        self.assertRaises(ValueError, trait_obj.ids_from_names,
                          self.ctx, [])

    def test_trait_create(self):
        t = trait_obj.Trait(self.ctx)
        t.name = 'CUSTOM_TRAIT_A'
        t.create()
        self.assertIsNotNone(t.id)
        self.assertEqual(t.name, 'CUSTOM_TRAIT_A')

    def test_trait_create_with_id_set(self):
        t = trait_obj.Trait(self.ctx)
        t.name = 'CUSTOM_TRAIT_A'
        t.id = 1
        self.assertRaises(exception.ObjectActionError, t.create)

    def test_trait_create_without_name_set(self):
        t = trait_obj.Trait(self.ctx)
        self.assertRaises(exception.ObjectActionError, t.create)

    def test_trait_create_duplicated_trait(self):
        trait = trait_obj.Trait(self.ctx)
        trait.name = 'CUSTOM_TRAIT_A'
        trait.create()
        tmp_trait = trait_obj.Trait.get_by_name(self.ctx, 'CUSTOM_TRAIT_A')
        self.assertEqual('CUSTOM_TRAIT_A', tmp_trait.name)
        duplicated_trait = trait_obj.Trait(self.ctx)
        duplicated_trait.name = 'CUSTOM_TRAIT_A'
        self.assertRaises(exception.TraitExists, duplicated_trait.create)

    def test_trait_get(self):
        t = trait_obj.Trait(self.ctx)
        t.name = 'CUSTOM_TRAIT_A'
        t.create()
        t = trait_obj.Trait.get_by_name(self.ctx, 'CUSTOM_TRAIT_A')
        self.assertEqual(t.name, 'CUSTOM_TRAIT_A')

    def test_trait_get_non_existed_trait(self):
        self.assertRaises(
            exception.TraitNotFound,
            trait_obj.Trait.get_by_name, self.ctx, 'CUSTOM_TRAIT_A')

    def test_bug_1760322(self):
        # Under bug # #1760322, if the first hit to the traits table resulted
        # in an exception, the sync transaction rolled back and the table
        # stayed empty; but _TRAITS_SYNCED got set to True, so it didn't resync
        # next time.
        # NOTE(cdent): With change Ic87518948ed5bf4ab79f9819cd94714e350ce265
        # syncing is no longer done in the same way, so the bug fix that this
        # test was testing is gone, but this test has been left in place to
        # make sure we still get behavior we expect.
        try:
            trait_obj.Trait.get_by_name(self.ctx, 'CUSTOM_GOLD')
        except exception.TraitNotFound:
            pass
        # Under bug #1760322, this raised TraitNotFound.
        trait_obj.Trait.get_by_name(self.ctx, os_traits.HW_CPU_X86_AVX2)

    def test_trait_destroy(self):
        t = trait_obj.Trait(self.ctx)
        t.name = 'CUSTOM_TRAIT_A'
        t.create()
        t = trait_obj.Trait.get_by_name(self.ctx, 'CUSTOM_TRAIT_A')
        self.assertEqual(t.name, 'CUSTOM_TRAIT_A')
        t.destroy()
        self.assertRaises(exception.TraitNotFound, trait_obj.Trait.get_by_name,
                          self.ctx, 'CUSTOM_TRAIT_A')

    def test_trait_destroy_with_standard_trait(self):
        t = trait_obj.Trait(self.ctx)
        t.id = 1
        t.name = 'HW_CPU_X86_AVX'
        self.assertRaises(exception.TraitCannotDeleteStandard, t.destroy)

    def test_traits_get_all(self):
        trait_names = ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C']
        for name in trait_names:
            t = trait_obj.Trait(self.ctx)
            t.name = name
            t.create()

        self._assert_traits_in(trait_names,
                               trait_obj.get_all(self.ctx))

    def test_traits_get_all_with_name_in_filter(self):
        trait_names = ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C']
        for name in trait_names:
            t = trait_obj.Trait(self.ctx)
            t.name = name
            t.create()

        traits = trait_obj.get_all(
            self.ctx,
            filters={'name_in': ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B']})
        self._assert_traits(['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B'], traits)

    def test_traits_get_all_with_non_existed_name(self):
        traits = trait_obj.get_all(
            self.ctx,
            filters={'name_in': ['CUSTOM_TRAIT_X', 'CUSTOM_TRAIT_Y']})
        self.assertEqual(0, len(traits))

    def test_traits_get_all_with_prefix_filter(self):
        trait_names = ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C']
        for name in trait_names:
            t = trait_obj.Trait(self.ctx)
            t.name = name
            t.create()

        traits = trait_obj.get_all(self.ctx, filters={'prefix': 'CUSTOM'})
        self._assert_traits(
            ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C'],
            traits)

    def test_traits_get_all_with_non_existed_prefix(self):
        traits = trait_obj.get_all(self.ctx, filters={"prefix": "NOT_EXISTED"})
        self.assertEqual(0, len(traits))

    def test_trait_delete_in_use(self):
        rp = self._create_provider('fake_resource_provider')
        t, = tb.set_traits(rp, 'CUSTOM_TRAIT_A')
        self.assertRaises(exception.TraitInUse, t.destroy)

    def test_traits_get_all_with_associated_true(self):
        rp1 = self._create_provider('fake_resource_provider1')
        rp2 = self._create_provider('fake_resource_provider2')
        trait_names = ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C']
        for name in trait_names:
            t = trait_obj.Trait(self.ctx)
            t.name = name
            t.create()

        associated_traits = trait_obj.get_all(
            self.ctx,
            filters={'name_in': ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B']})
        rp1.set_traits(associated_traits)
        rp2.set_traits(associated_traits)
        self._assert_traits(
            ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B'],
            trait_obj.get_all(self.ctx, filters={'associated': True}))

    def test_traits_get_all_with_associated_false(self):
        rp1 = self._create_provider('fake_resource_provider1')
        rp2 = self._create_provider('fake_resource_provider2')
        trait_names = ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B', 'CUSTOM_TRAIT_C']
        for name in trait_names:
            t = trait_obj.Trait(self.ctx)
            t.name = name
            t.create()

        associated_traits = trait_obj.get_all(
            self.ctx,
            filters={'name_in': ['CUSTOM_TRAIT_A', 'CUSTOM_TRAIT_B']})
        rp1.set_traits(associated_traits)
        rp2.set_traits(associated_traits)
        self._assert_traits_in(
            ['CUSTOM_TRAIT_C'],
            trait_obj.get_all(self.ctx, filters={'associated': False}))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Reshaper schema for Placement API."""

import copy

from placement.schemas import allocation
from placement.schemas import common
from placement.schemas import inventory


ALLOCATIONS = copy.deepcopy(allocation.POST_ALLOCATIONS_V1_28)
# In the reshaper we need to allow allocations to be an empty dict
# because it may be the case that there simply are no allocations
# (now) for any of the inventory being moved.
ALLOCATIONS['minProperties'] = 0
POST_RESHAPER_SCHEMA = {
    "type": "object",
    "properties": {
        "inventories": {
            "type": "object",
            "patternProperties": {
                # resource provider uuid
                common.UUID_PATTERN: inventory.PUT_INVENTORY_SCHEMA,
            },
            # We expect at least one inventories, otherwise there is no reason
            # to call the reshaper.
            "minProperties": 1,
            "additionalProperties": False,
        },
        "allocations": ALLOCATIONS,
    },
    "required": [
        "inventories",
        "allocations",
    ],
    "additionalProperties": False,
}

POST_RESHAPER_SCHEMA_V1_34 = copy.deepcopy(POST_RESHAPER_SCHEMA)
ALLOCATIONS_V1_34 = copy.deepcopy(allocation.POST_ALLOCATIONS_V1_34)
ALLOCATIONS_V1_34['minProperties'] = 0
POST_RESHAPER_SCHEMA_V1_34['properties']['allocations'] = ALLOCATIONS_V1_34

POST_RESHAPER_SCHEMA_V1_38 = copy.deepcopy(POST_RESHAPER_SCHEMA_V1_34)
ALLOCATIONS_V1_38 = copy.deepcopy(allocation.POST_ALLOCATIONS_V1_38)
ALLOCATIONS_V1_38['minProperties'] = 0
POST_RESHAPER_SCHEMA_V1_38['properties']['allocations'] = ALLOCATIONS_V1_38



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections

from placement import direct
from placement.tests.functional import base
from placement.tests.functional.db import test_base as tb


class TestWideTreeAllocationCandidateExplosion(base.TestCase):
    """Test candidate generation ordering and limiting in wide symmetric trees,
     i.e. with trees of many similar child RPs.
    """

    def setUp(self):
        super().setUp()
        self.headers = {
            'x-auth-token': 'admin',
            'content-type': 'application/json',
            'OpenStack-API-Version': 'placement 1.38',
            'X_ROLES': 'admin,service'
        }

        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 100000, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")

    def create_tree(self, num_roots, num_child, num_res_per_child):
        self.roots = {}

        for i in range(num_roots):
            compute = tb.create_provider(
                self.context, f'compute{i}')
            self.roots[compute.uuid] = compute.name
            tb.add_inventory(compute, 'VCPU', 8)
            tb.add_inventory(compute, 'MEMORY_MB', 4096)
            tb.add_inventory(compute, 'DISK_GB', 500)

            for j in range(num_child):
                child = tb.create_provider(
                    self.context, f'compute{i}:PF{j}', parent=compute.uuid)
                tb.add_inventory(child, 'CUSTOM_VF', num_res_per_child)

    @staticmethod
    def get_candidate_query(num_groups, num_res, limit):
        query = ("/allocation_candidates?"
                 "resources=DISK_GB%3A20%2CMEMORY_MB%3A2048%2CVCPU%3A2")

        for g in range(num_groups):
            query += f"&resources{g}=CUSTOM_VF%3A{num_res}"

        query += "&group_policy=none"
        query += f"&limit={limit}"

        return query

    def _test_num_candidates_and_computes(
        self, computes, pfs, vfs_per_pf, req_groups, req_res_per_group,
        req_limit, expected_candidates, expected_computes_with_candidates
    ):
        self.create_tree(
            num_roots=computes, num_child=pfs, num_res_per_child=vfs_per_pf)

        conf = self.conf_fixture.conf
        with direct.PlacementDirect(conf) as client:
            resp = client.get(
                self.get_candidate_query(
                    num_groups=req_groups, num_res=req_res_per_group,
                    limit=req_limit),
                headers=self.headers)
            self.assertEqual(200, resp.status_code)

        body = resp.json()
        self.assertEqual(expected_candidates, len(body["allocation_requests"]))

        root_rps = set(self.roots.keys())
        roots_with_candidates = set()
        nr_of_candidates_per_compute = collections.Counter()
        for ar in body["allocation_requests"]:
            allocated_rps = set(ar["allocations"].keys())
            root_allocated_rps = allocated_rps.intersection(root_rps)
            roots_with_candidates |= root_allocated_rps
            nr_of_candidates_per_compute.update(root_allocated_rps)

        self.assertEqual(
            expected_computes_with_candidates, len(roots_with_candidates))

    def test_all_candidates_generated_and_returned(self):
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=2, req_res_per_group=1,
            req_limit=1000,
            expected_candidates=2 * 64, expected_computes_with_candidates=2,)

    def test_requested_limit_is_hit_result_balanced(self):
        # 8192 possible candidates, all generated, returned 1000,
        # result is balanced due to python sets usage
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=4, req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=2)

    def test_too_many_candidates_global_limit_is_hit_result_unbalanced(self):
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "depth-first",
            group="placement")
        # With max_allocation_candidates set to 100k limit this test now
        # runs in reasonable time (10 sec on my machine), without that it would
        # time out.
        # However, with depth-first strategy and with the global limit in place
        # only the first compute gets candidates.
        # 524288 valid candidates, the generation stops at 100k candidates,
        # only 1000 is returned, result is unbalanced as the first 100k
        # candidate is always from the first compute.
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=6, req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=1)

    def test_too_many_candidates_global_limit_is_hit_breadth_first_balanced(
        self
    ):
        # With max_allocation_candidates set to 100k limit this test now
        # runs in reasonable time (10 sec on my machine), without that it would
        # time out.
        # With the round-robin candidate generator in place the 100k generated
        # candidates spread across both computes now.
        # 524288 valid candidates, the generation stops at 100k candidates,
        # only 1000 is returned, result is balanced between the computes
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=6, req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=2)

    def test_global_limit_hit(self):
        # 8192 possible candidates, global limit is set to 8000, higher request
        # limit so number of candidates are limited by the global limit
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 8000, group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=4, req_res_per_group=1,
            req_limit=9000,
            expected_candidates=8000, expected_computes_with_candidates=2)

    def test_no_global_limit(self):
        # 8192 possible candidates, there is no global limit, high request
        # limit so all candidates returned
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", -1, group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=8, vfs_per_pf=8, req_groups=4, req_res_per_group=1,
            req_limit=9000,
            expected_candidates=8192, expected_computes_with_candidates=2)

    def test_breadth_first_strategy_generates_stable_ordering(self):
        """Run the same query twice against the same two tree and assert that
        response text is exactly the same proving that even with breadth-first
        strategy the candidate ordering is stable.
        """

        self.create_tree(num_roots=2, num_child=8, num_res_per_child=8)

        def query():
            return client.get(
                self.get_candidate_query(
                    num_groups=2, num_res=1,
                    limit=1000),
                headers=self.headers)

        conf = self.conf_fixture.conf
        with direct.PlacementDirect(conf) as client:
            resp = query()
            self.assertEqual(200, resp.status_code)
            body1 = resp.text

            resp = query()
            self.assertEqual(200, resp.status_code)
            body2 = resp.text

            self.assertEqual(body1, body2)

    def test_many_non_viable_candidates_8_8(self):
        # This simulates that we have a single resource per RP (in this case
        # one VF, but it could be on PF resource it does not matter). We have
        # many RPs and we request many groups of one resource. This creates a
        # situation where even if the number of candidates are limited by
        # max_allocation_candidates the possible number of candidates generated
        # by satisfying each group independently and then generating all
        # possible combinations results in an exponential number of possible
        # candidate from which most of them are invalid due to two groups
        # independently satisfied by the same single resource.
        # Filtering this list for valid candidates take too much time.
        #
        # We have 8 RPs with 1 resource, and we request 8 groups with
        # 1 resource.
        # The full candidate matrix by satisfying each group independently
        # (G is request group, R is RP):
        #
        # G1: [R1, R2,..., R8]
        # G2: [R1, R2,..., R8]
        # ...
        # G8: [R1, R2,..., R8]
        #
        # Creating all the possible combinations and checking if they are
        # valid (C is candidate, G1-R1 means G1 group satisfied from R1 RP):
        # C1: [G1-R1, G2-R1, ..., G8-R1] # invalid R1 has 1 res but C1 needs 8
        # C2: [G1-R1, G2-R1, ..., G8-R2] # invalid R1 has 1 res but C2 needs 7
        # ...
        # Cx: [G1-R1, G2-R2, ..., G8-R8] # valid each Rx has 1 res and
        #                                # Cx ask form 1 res each
        #
        # After bugfix #2126751 placement is changed not to generate all these
        # candidate, but instead if it finds that a candidate is invalid
        # because a prefix of the groups (G1-R1 and G2-R2) causing an
        # overallocation then all possible candidates that starts with
        # the same prefix are removed from the search space. This moves the
        # algorithm from exponential to factorial.
        #
        # This runs in 1.2 seconds. If you bump this from 1000 to 10k maximum
        # candidates then it will run in 106 seconds.
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self._test_num_candidates_and_computes(
            computes=1, pfs=8, vfs_per_pf=1, req_groups=8, req_res_per_group=1,
            req_limit=10000,
            expected_candidates=1000, expected_computes_with_candidates=1)

    def test_many_non_viable_candidates_21_8(self):
        # This is runs in 0.14 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self._test_num_candidates_and_computes(
            computes=1, pfs=21, vfs_per_pf=1, req_groups=8,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=1)

    def test_many_non_viable_candidates_21_16(self):
        # This is runs in 0.21 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self._test_num_candidates_and_computes(
            computes=1, pfs=21, vfs_per_pf=1, req_groups=16,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=1)

    def test_many_non_viable_candidates_21_21(self):
        # This is runs in 3 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self._test_num_candidates_and_computes(
            computes=1, pfs=21, vfs_per_pf=1, req_groups=21,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=1)

    def test_many_non_viable_candidates_21_8_two_computes(self):
        # This is runs in 0.17 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=21, vfs_per_pf=1, req_groups=8,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=2)

    def test_many_non_viable_candidates_21_21_two_computes(self):
        # This is runs in 1.6 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=21, vfs_per_pf=1, req_groups=21,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=2)

    def test_many_non_viable_candidates_32_32_two_computes(self):
        # This is runs in 2.45 seconds
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 1000, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=32, vfs_per_pf=1, req_groups=32,
            req_res_per_group=1,
            req_limit=1000,
            expected_candidates=1000, expected_computes_with_candidates=2)

    def test_many_non_viable_candidates_48_48_two_computes(self):
        # This is runs in 0.36 seconds with 100 max candidates and runs in
        # 3.6 seconds for 1000.
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 100, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=48, vfs_per_pf=1, req_groups=48,
            req_res_per_group=1,
            req_limit=100,
            expected_candidates=100, expected_computes_with_candidates=2)

    def test_many_non_viable_candidates_64_64_two_computes(self):
        # This is runs in 0.35 seconds for 10 max candidates and runs in
        # 0.5 seconds with 100 and runs in 5.1 seconds with 1000.
        self.conf_fixture.conf.set_override(
            "optimize_for_wide_provider_trees", True, group="workarounds")
        self.conf_fixture.conf.set_override(
            "max_allocation_candidates", 10, group="placement")
        self.conf_fixture.conf.set_override(
            "allocation_candidates_generation_strategy", "breadth-first",
            group="placement")
        self._test_num_candidates_and_computes(
            computes=2, pfs=64, vfs_per_pf=1, req_groups=64,
            req_res_per_group=1,
            req_limit=10,
            expected_candidates=10, expected_computes_with_candidates=2)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fixtures for Placement tests."""


from oslo_config import cfg
from oslo_db.sqlalchemy import test_fixtures

from placement.db.sqlalchemy import migration
from placement import db_api as placement_db
from placement import deploy
from placement.objects import resource_class
from placement.objects import trait


class Database(test_fixtures.GeneratesSchema, test_fixtures.AdHocDbFixture):
    def __init__(self, conf_fixture, set_config=False):
        """Create a database fixture."""
        super(Database, self).__init__()
        if set_config:
            try:
                conf_fixture.register_opt(
                    cfg.StrOpt('connection'), group='placement_database')
            except cfg.DuplicateOptError:
                # already registered
                pass
            conf_fixture.config(connection='sqlite://',
                                group='placement_database')
        self.conf_fixture = conf_fixture
        self.get_engine = placement_db.get_placement_engine
        placement_db.configure(self.conf_fixture.conf)

    def get_enginefacade(self):
        return placement_db.placement_context_manager

    def generate_schema_create_all(self, engine):
        # note: at this point in oslo_db's fixtures, the incoming
        # Engine has **not** been associated with the global
        # context manager yet.
        migration.create_schema(engine)

        # so, to work around that placement's setup code really wants to
        # use the enginefacade, we will patch the engine into it early.
        # oslo_db is going to patch it anyway later.  So the bug in oslo.db
        # is that code these days really wants the facade to be set up fully
        # when it's time to create the database.  When oslo_db's fixtures
        # were written, enginefacade was not in use yet so it was not
        # anticipated that everyone would be doing things this way
        _reset_facade = placement_db.placement_context_manager.patch_engine(
            engine)
        self.addCleanup(_reset_facade)

        # Make sure db flags are correct at both the start and finish
        # of the test.
        self.addCleanup(self.cleanup)
        self.cleanup()

        # Sync traits and resource classes.
        deploy.update_database(self.conf_fixture.conf)

    def cleanup(self):
        trait._TRAITS_SYNCED = False
        resource_class._RESOURCE_CLASSES_SYNCED = False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Block on null root_provider_id

Revision ID: 611cd6dffd7b
Revises: b4ed3a175331
Create Date: 2019-05-09 13:57:04.874293

"""
from alembic import context
import sqlalchemy as sa
from sqlalchemy import func as sqlfunc
from sqlalchemy import MetaData, Table, select

# revision identifiers, used by Alembic.
revision = '611cd6dffd7b'
down_revision = 'b4ed3a175331'
branch_labels = None
depends_on = None


def upgrade():
    connection = context.get_bind()

    meta = MetaData()
    meta.reflect(bind=connection)
    resource_providers = Table(
        'resource_providers',
        meta,
        autoload_with=connection,
    )

    query = select(
        sqlfunc.count(),
    ).select_from(
        resource_providers,
    ).where(
        resource_providers.c.root_provider_id == sa.null()
    )

    if connection.scalar(query):
        raise Exception('There is at least one resource provider table '
                        'record which is missing its root provider id. '
                        'Run the "placement-manage db '
                        'online_data_migrations" command.')
